#------------------------------------------------------------------
# $Copyright: Copyright (c) 2019 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/OracleApps/OracleApps.pm,v 1.32 2015/01/22 12:45:32 mrampura Exp $
#
# $Author: mrampura $
# $Date: 2015/01/22 12:45:32 $
# $Revision: 1.32 $
#
# Provides:                            OracleApps
# Author:                              aprabhu1
# Creation Date:                       23 Feb, 2006
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package OracleApps;

my ( $CLUS_HOME, $sPrefix ) = ();

#---------------------
# Setup VCS/VAD home..
#---------------------
if ( $ENV { CLUSTER_HOME } ) {
  $CLUS_HOME ||= $ENV { VAD_HOME } ||= "/opt/VRTSvad";
  $sPrefix = 'vad';
}
else { # VCS
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
  $sPrefix = 'vcs';
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = "$CLUS_HOME/bin/hares";
my $sCmdHatype = "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use SystemD;        # Support generalized SystemD Functions
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses
use ag_i18n_inc;

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use POSIX qw ( uname );
use POSIX qw ( strftime );
use File::Path qw ( mkpath );
use File::Basename;
use Cwd;

$::gnOracleAppsVersion = undef;
@::gsLogString = ();
%::hOracleAppsTypeValues = ();
my $sFormsMode = ();

my @MONITOR_PROC_LIST = ();
my @VALID_SERVERTYPE_LIST        = ();
my %hScripts                     = ();
my @FORMSSERVERPROC_LIST         = ();
my @FORMSMETRICSSERVERPROC_LIST  = ();
my @FORMSMETRICSCLIENTPROC_LIST  = ();
my @REPORTSERVERPROC_LIST        = ();
my @DISCOVERERPROC_LIST          = ();
my @LISTENERPROC_LIST            = ();
my @FULFILLMENTSERVERPROC_LIST   = ();
my @OAFMPROC_LIST                = ();
my @OACOREPROC_LIST              = ();
my @WEBSERVERPROC_LIST           = ();
my @OPMNPROC_LIST                = ();
my @ADMINSERVERPROC_LIST         = ();
my @NODEMANAGERPROC_LIST         = ();

use constant EPOP => 3;

#--------------------------------
# Define OracleApps subroutines..
#--------------------------------

=pod

#---------------------------------------------------------------
 Function : void GetFormsMode ( IN String User,
                                IN String ScriptHome,
                                IN String EnvFile,
                                IN Integer Port );

 Description : Set the Servertype, Script names, and Processes
               Lists for OracleApps componants based on the
               OracleApps Version.

 Parameters :

 $sUser                     The User to start/stop the Component
 $sScriptHome               Pathname of the Scripts directory
 $sEnvFile                  Full pathname to the env file to setup the env
 $iPort                     Port Number the component is listening on

 Returns : String value in the set of {servlet/socket} as
           servlet -> Forms running in Servlet Mode
           socket  -> Forms running in Socket  Mode
#---------------------------------------------------------------

=cut

sub GetFormsMode
{
  my ( $sUser,
       $sScriptHome,
       $sEnvFile,
       $iPort ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------

  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  User                     [%s]
  ScriptHome               [%s]
  EnvFile                  [%s]
  Port                     [%s]",
  $sMyName, $sUser, $sScriptHome, $sEnvFile, $iPort);

  if ( $iPort && $::gnOracleAppsVersion eq "12" )
  {
    &logTrace ( 20002, "<%s> Forms Running in Socket Mode", $sMyName );
    return "socket";
  }

  my $iExitCode = 0;
  my @lsCommandOut = ();

  my $sInstHome = "$sScriptHome/../..";
  if ( ! -d $sInstHome )
  {
       $iExitCode = 1; # DiskGroup is not mounted or ScriptHome is invalid
  }
  else
  {
      my $sContextFile = $sInstHome . "/appl/admin/" . basename(Cwd::abs_path($sInstHome)) . ".xml";
      open(FH, $sContextFile);
      my @buf = <FH>;
      close(FH);
      my @lines = grep (/s_frmConnectMode/ && /socket/, @buf);
      if (@lines)
      {
            $iExitCode = 0;
      }
      else
      {
            $iExitCode = 1;
      }
  }
  #---------------------------------------------
  # Exit Code = 0 -> Servlet Mode
  #           = 1 -> Socket  Mode
  #---------------------------------------------
  if ( $iExitCode )
  {
    &logTrace ( 20003, "<%s> Forms Running in Servlet Mode" ,$sMyName );
    return "servlet";
  }
  else
  {
    &logTrace ( 20004, "<%s> Forms Running in Socket Mode", $sMyName );
    return "socket";
  }
}

=pod

#---------------------------------------------------------------
 Function : void SetValues ( IN String FormsMode
			     IN String OracleHome
                 IN String ServerType
                 IN String ServerName);

 Description : Set the Servertype, Script names, and Processes
               Lists for OracleApps componants based on the 
               OracleApps Version.

 Parameters :

 $sFormsMode    	Forms Configured mode i.e. (Server/Socket)
 $sOracleHome		Pathname of the ORACLE_HOME directory
 $sServerType       Oracle Application Component
 $sServerName       Name of Oracle Application Component 

 Returns : NONE
#---------------------------------------------------------------

=cut

sub SetValues
{
   my( $sFormsMode, $sOracleHome, $sServerType, $sServerName, $iPort ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20005, "Subroutine <%s> called with:
  FormsMode                [%s]
  OracleHome		   [%s] 
  ServerType            [%s]
  ServerName            [%s]
  Port                  [%s]",
  $sMyName, $sFormsMode, $sOracleHome, $sServerType, $sServerName, $iPort);
  $sServerType =~ tr/A-Z/a-z/;
    if ($::gnOracleAppsVersion eq "11" ) {
        #-------------------------------------------
        # Define test version 11 Constants ..
        #-------------------------------------------
    
        @VALID_SERVERTYPE_LIST   = ( "formsserver",
                                     "formsmetricsserver",
                                     "formsmetricsclient",
                                     "reportserver",
                                     "discoverer",
                                     "listener",
                                     "fulfillmentserver",
                                     "webserver" );

        %hScripts = ( "formsserver"        => "adfrmctl.sh",
                      "formsmetricsserver" => "adfmsctl.sh",
                      "formsmetricsclient" => "adfmcctl.sh",
                      "reportserver"       => "adrepctl.sh",
                      "discoverer"         => "addisctl.sh",
                      "listener"           => "adalnctl.sh",
                      "fulfillmentserver"  => "jtffmctl.sh",
                      "webserver"          => "adapcctl.sh");

        @FORMSSERVERPROC_LIST          = ( "f60srvm", "f60webmx" );
        @FORMSMETRICSSERVERPROC_LIST   = ( "d2ls60" );
        @FORMSMETRICSCLIENTPROC_LIST   = ( "d2lc60" );
        @REPORTSERVERPROC_LIST         = ( "rwmts60" );
        @DISCOVERERPROC_LIST           = ( "osagent" , "oad");
        @LISTENERPROC_LIST             = ( "$sOracleHome/bin/tnslsnr" );
        @WEBSERVERPROC_LIST            = ( "httpd" );
        @FULFILLMENTSERVERPROC_LIST    = ( "java.*FulfillmentServer" );
    }
    elsif ($::gnOracleAppsVersion eq "12" ) {
        #-------------------------------------------
        # Define test version 12 Constants ..
        #-------------------------------------------
    
        @VALID_SERVERTYPE_LIST   = ( "opmn", 
                                     "formsserver",
                                     "oafm",
                                     "oacore",
                                     "listener",
                                     "fulfillmentserver",
                                     "webserver" );
 
        my $sFormsScript = undef;
        if ( $sFormsMode eq "socket" )
        {
            $sFormsScript = "adformsrvctl.sh";
            @FORMSSERVERPROC_LIST = ( "frmsrv", "frmweb" );
        }
        else
        {
            $sFormsScript = "adformsctl.sh";
            @FORMSSERVERPROC_LIST = ( "-Doracle.home=$sOracleHome\\s+.*-Doracle.oc4j.instancename=forms\\b" );
        }
    
        %hScripts = ( "opmn"               => "adopmnctl.sh",
                      "formsserver"        => "$sFormsScript",
                      "oafm"               => "adoafmctl.sh",
                      "oacore"             => "adoacorectl.sh",
                      "listener"           => "adalnctl.sh",
                      "fulfillmentserver"  => "jtffmctl.sh",
                      "discoverer"         => "opmnctl.sh",
                      "webserver"          => "adapcctl.sh");
    
        @OPMNPROC_LIST                 = ( "$sOracleHome/opmn/bin/opmn" );
        @OAFMPROC_LIST                 = ( "-Doracle.home=$sOracleHome\\s+.*-Doracle.oc4j.instancename=oafm\\b" );
        @OACOREPROC_LIST               = ( "-Doracle.home=$sOracleHome\\s+.*-Doracle.oc4j.instancename=oacore\\b" );
        @LISTENERPROC_LIST             = ( "$sOracleHome/bin/tnslsnr" );
        @FULFILLMENTSERVERPROC_LIST    = ( "java\\s+.*-Doracle.home=$sOracleHome\\s+.*"."FulfillmentServer" );
        @WEBSERVERPROC_LIST            = ( "$sOracleHome/Apache/Apache/bin/httpd", "$sOracleHome/Apache/Apache/bin/rotatelogs" );
        @DISCOVERERPROC_LIST           = ( "osagent" , "oad");
    }
    else {
        #-------------------------------------------
        # Define test version 12.2 Constants ..
        #-------------------------------------------
        @VALID_SERVERTYPE_LIST   = ( "opmn",
                                     "formsserver",
                                     "oafm",
                                     "oacore",
                                     "listener",
                                     "fulfillmentserver",
                                     "webserver",
			             "adminserver",
                                     "nodemanager" );

        my $sFormsScript = undef;
        if ( $sFormsMode eq "socket" )
        {
            $sFormsScript = "adformsrvctl.sh";
            @FORMSSERVERPROC_LIST = ( "frmsrv", "frmweb" );
        }
        else
        {
            $sFormsScript = "admanagedsrvctl.sh";
            # @FORMSSERVERPROC_LIST = ( "-Doracle.home=$sOracleHome\\s+.*-Doracle.oc4j.instancename=forms\\b" );
        }

        %hScripts = ( "opmn"               => "adopmnctl.sh",
                      "formsserver"        => "$sFormsScript",
                      "forms"              => "$sFormsScript",
                      "oafm"               => "admanagedsrvctl.sh",
                      "oacore"             => "admanagedsrvctl.sh",
                      "listener"           => "adalnctl.sh",
                      "fulfillmentserver"  => "jtffmctl.sh",
                      "discoverer"         => "opmnctl.sh",
                      "webserver"          => "adapcctl.sh",
	              "adminserver"        => "adadminsrvctl.sh",
                      "nodemanager"        => "adnodemgrctl.sh");
        
        my $sOSType = Arch::GetOSType();
        
        @OPMNPROC_LIST                 = ( "$sOracleHome/opmn/bin/opmn" );
        @OAFMPROC_LIST                 = ( "-Doracle.home=$sOracleHome\\s+.*-Doracle.oc4j.instancename=oafm\\b" );
        @OACOREPROC_LIST               = ( "-Doracle.home=$sOracleHome\\s+.*-Doracle.oc4j.instancename=oacore\\b" );
        @LISTENERPROC_LIST             = ( "$sOracleHome/bin/tnslsnr" );
        @FULFILLMENTSERVERPROC_LIST    = ( "java\\s+.*-Doracle.home=$sOracleHome\\s+.*"."FulfillmentServer" );
        @WEBSERVERPROC_LIST            = ( "$sOracleHome/.*/bin/httpd", "$sOracleHome/.*/bin/rotatelogs" );
        @DISCOVERERPROC_LIST           = ( "osagent" , "oad");
	@ADMINSERVERPROC_LIST          = ( "java\\s+.*-Dweblogic.Name=AdminServer");
        @NODEMANAGERPROC_LIST          = ( "java\\s+.*weblogic.NodeManager");

        if ( $sServerType eq "oacore" || $sServerType eq "oafm" || ( $sServerType eq "formsserver" && $sFormsMode eq "servlet") ) {

            @OAFMPROC_LIST = "java.*(management.server=\\w+://.*$iPort.*Name=$sServerName\\b)|(Name=$sServerName\\b.*management.server=\\w+://.*$iPort)";
            @OACOREPROC_LIST = "java.*(management.server=\\w+://.*$iPort.*Name=$sServerName\\b)|(Name=$sServerName\\b.*management.server=\\w+://.*$iPort)";
            @FORMSSERVERPROC_LIST = "java.*(management.server=\\w+://.*$iPort.*Name=$sServerName\\b)|(Name=$sServerName\\b.*management.server=\\w+://.*$iPort)";
        
        }
    }
}


=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String User,
                              IN String OracleAppsVersion,
                              IN String OracleHome,
                              IN String ScriptHome,
                              IN String EnvFile,
                              IN String ServerType,
                              IN String ServerName,
                              IN String WLSPassword,
                              IN String AppsPassword,
                              IN Integer Port,
                              IN String DBConnectString,
                              IN Integer SecondLevelMonitor,
                              IN Integer MonitorProgram,
                              IN String VirtualHostname );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localUser                     The User to start/stop the Concurrent Manager
 $localOracleAppsVersion        Oracle Applictions Version (11,12 or 12.2)
 $localOracleHome               Pathname of the ORACLE_HOME directory
 $localScriptHome               Pathname of the Scripts directory
 $localEnvFile                  Full pathname to the env file to setup the env
 $localServerType               Oracle Application Component
 $localServerName               Name of Oracle Application Component
 $localWLSPassword              Password of Weblogic Application Server
 $localAppsPassword             Password for Apps schema
 $localPort                     Port Number the component is listening on
 $localDBConnectString          The TNS connect string to ping the listener
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localMonitorProgram           Full path to external monitor program
 $localVirtualHostname          Virtual hostname associated with this CM instance
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localUser,
       $localOracleAppsVersion,
       $localOracleHome,
       $localScriptHome,
       $localEnvFile,
       $localServerType,
       $localServerName,
       $localWLSPassword,
       $localAppsPassword,
       $localPort,
       $localDBConnectString,
       $localSecondLevelMonitor,
       $localMonitorProgram,
       $localVirtualHostname ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20006, "Subroutine <%s> called with:
  EntryPointName           [%s]
  User                     [%s]
  OracleAppsVersion        [%s]
  OracleHome               [%s]
  ScriptHome               [%s]
  EnvFile                  [%s]
  ServerType               [%s]
  ServerName               [%s]
  WLSPassword              [%s]
  AppsPassword             [%s]
  Port                     [%s]
  DBConnectString          [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]
  VirtualHostname          [%s]",

  $sMyName, $sEntryPointName, $localUser, $localOracleAppsVersion ,
  $localOracleHome, $localScriptHome, $localEnvFile, $localServerType,
  $localServerName, $localWLSPassword, $localAppsPassword, $localPort, $localDBConnectString,
  $localSecondLevelMonitor, $localMonitorProgram, $localVirtualHostname );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName  User  OracleAppsVersion OracleHome ScriptHome 
       EnvFile ServerType ServerName WLSPassword  AppsPassword Port DBConnectString 
       SecondLevelMonitor MonitorProgram VirtualHostname );

  @hsArgList { @lsArgs } = @_; # Assign..

  #--------------------------------------------------
  # Check if OracleAppsVersion is missing or invalid
  #--------------------------------------------------

  if ( "11" eq $localOracleAppsVersion ||  "12" eq $localOracleAppsVersion || "12.2" eq $localOracleAppsVersion)
  {
     $sFormsMode = "servlet";  # Default mode is servlet 
     if ( "11" ne $localOracleAppsVersion && $localServerType eq "FormsServer" )
     {
        $sFormsMode = GetFormsMode( $localUser, $localScriptHome, $localEnvFile, $localPort );
     }
     OracleApps::SetValues( $sFormsMode, $localOracleHome, $localServerType, $localServerName, $localPort ); # Set the Version specific constant values.
  }
  else
  {
     &logWarn ( 20007, "<%s> Invalid Oracle Applications E-Business Suite Version : <%s>",
                         $sMyName, $localOracleAppsVersion );
     $iReturnCode = 0;
  }

  $iReturnCode || return $iReturnCode; # Return, if Version not validated

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------

  my @lsOnlineArgs = ();
  my @lsOfflineArgs = ();
  my @lsMonitorArgs = ();
  my @lsCleanArgs   = ();

  if ( "11" eq $localOracleAppsVersion )
  {
     @lsOnlineArgs  = qw ( User OracleHome ScriptHome EnvFile ServerType Port VirtualHostname );
     @lsOfflineArgs = qw ( User OracleHome ScriptHome EnvFile ServerType VirtualHostname);
     @lsMonitorArgs = qw ( User OracleHome ServerType Port VirtualHostname);
     @lsCleanArgs   = qw ( User ScriptHome EnvFile ServerType VirtualHostname);
  }
  else
  {
     @lsOnlineArgs  = qw ( User OracleHome ScriptHome ServerType VirtualHostname );
     @lsOfflineArgs = qw ( User OracleHome ScriptHome ServerType VirtualHostname);
     @lsMonitorArgs = qw ( User OracleHome ServerType VirtualHostname);
     @lsCleanArgs   = qw ( User ScriptHome ServerType VirtualHostname);
  }

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{$hlsNeeded { $sEntryPointName }};

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20008,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------
  # Individual validation starts..
  #-------------------------------

  if ( $sEntryPointName eq "online"  ||
       $sEntryPointName eq "offline" )
  {
     if ( ! -d $localOracleHome )
     {
       &logError(20009,"<%s>:OracleHome [%s] is not accessible.",
                       $sMyName,$localOracleHome);
       $iReturnCode = 0;
     }
     if ( ! -d $localScriptHome )
     {
       &logError(20010,"<%s>:ScriptHome [%s] is not accessible.",
                       $sMyName,$localScriptHome);
       $iReturnCode = 0;
     }
     if ( 11 == $localOracleAppsVersion && ! -e $localEnvFile )
     {
       &logError(20011,"<%s>:EnvFile [%s] is not accessible.",
                       $sMyName,$localEnvFile);
       $iReturnCode = 0;
     }
  }

  $localServerType =~ tr/A-Z/a-z/;
  my $matchedServerType = 0;
  foreach my $inputServerType (@VALID_SERVERTYPE_LIST)
  {
     if ( $localServerType eq $inputServerType )
     {
         $matchedServerType = 1;
     }
  }
  if ( $matchedServerType eq 0 )
  {
     &logError(20012,"<%s>:ServerType [%s] is not valid server type.", 
                      $sMyName, $localServerType);
     $iReturnCode = 0;
  }
  
  if ( ( $localServerType eq "AdminServer" ) && ( ! defined $localAppsPassword  ) ) {
    &logError(20174,"<%s>: Apps Password is required for AdminServer component", 
                     $sMyName );
  }

  #--------------------------------------
  # MonitorProgram has to be executable..
  #--------------------------------------
  if ( $sEntryPointName eq "online"  )
  {
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) 
    {                        
        &logError ( 20013,
        "<%s> MonitorProgram [%s] not executable", $sMyName, $1 );
        $iReturnCode = 0; # Failure..
    }
  }


  $iReturnCode ||
  &logWarn ( 20014,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN Integer State,
                         IN Integer IState,
                         IN Integer ExpirationTime,
                         IN String  User,
                         IN String OracleHome,
                         IN String ScriptHome,
                         IN String EnvFile,
                         IN String ServerType,
                         IN String ServerName,
                         IN String WLSPassword,
                         IN String AppsPassword,
                         IN Integer Port,
                         IN String DBConnectString,
                         IN String MonitorProgram,
                         IN String VirtualHostname );

 Description : Start the Application COmponent defined by ServerType

 Parameters :

 $iState                    VCS State of the resource
 $iIState                   VCS Intermidiate State of the resource
 $iExpirationTime           Amount of time left to execute this procedure
 $sUser                     The User to start/stop the Concurrent Manager
 $sOracleHome               Pathname of the ORACLE_HOME directory
 $sScriptHome               Pathname of the Scripts directory
 $sEnvFile                  Full pathname to the env file to setup the env
 $sServerType               Oracle Application Component
 $sServerName               Name of Oracle Application Component
 $sWLSPassword              Password of Weblogic Application Server
 $sAppsPassword             Password for Apps schema
 $iPort                     Port used by this server
 $sDBConnectString          DB Connect string ( if component is listener)
 $sMonitorProgram           External MonitorProgram to run
 $VirtualHostname           Virtual hostname associated with this CM instance
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iState,
       $iIState,
       $iExpirationTime,
       $sUser,
       $sOracleHome,
       $sScriptHome,
       $sEnvFile,
       $sServerType,
       $sServerName,
       $sWLSPassword,
       $sAppsPassword,
       $iPort,
       $sDBConnectString,
       $sMonitorProgram,
       $sVirtualHostname,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20015, "Subroutine <%s> called with: ,
  State                    [%s]
  iState                   [%s]
  User                     [%s]
  OracleHome               [%s]
  ScriptHome               [%s]
  EnvFile                  [%s]
  ServerType               [%s]
  ServerName               [%s]
  WLSPassword              [%s]
  AppsPassword             [%s]
  Port                     [%s]
  DBConnectString          [%s]
  MonitorProgram           [%s]
  VirtualHostname          [%s]
  SecondLevelMonitor       [%s]",

  $sMyName, $iState, $iIState, $sUser,  $sOracleHome, $sScriptHome, $sEnvFile, 
  $sServerType, $sServerName, $sWLSPassword, $sAppsPassword, $iPort, $sDBConnectString, 
  $sMonitorProgram, $sVirtualHostname, $iSecondLevelMonitor );
  

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  $sServerType =~ tr/A-Z/a-z/;
  my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  #--------------------------------------
  # Remove lock file, if any            
  #--------------------------------------
  if ( "12" eq $::gnOracleAppsVersion )
  {
      my $sOC4JComponant = ""; 
      my @lsFiles = ();
      if ( $sServerType eq "formsserver" || $sServerType eq "oacore" || $sServerType eq "oafm" )
      {
          if ( $sServerType eq "formsserver" )
          {
              $sOC4JComponant = "forms";
          }
          else
          {
              $sOC4JComponant = $sServerType;
          }   
          my $sPerstDir = "$sScriptHome/../../ora/10.1.3/j2ee/" . $sOC4JComponant . "/persistence/" . $sOC4JComponant . "_default_group_1/";
          my @LockFileList = ( "jms.state.lock", 
                               "scheduler_jobstore.lock", 
                               "Oc4jJmsExceptionQueue.lock", 
                               "scheduler_notifications.lock", 
                               "scheduler_events.lock" );
   
          foreach my $sLockFile (@LockFileList)
          {
               my $sLockFileWithPath = $sPerstDir . "/" . $sLockFile;  
               if ( -e $sLockFileWithPath ) 
               {
                   &logTrace ( 20016,  "<%s> Located file [%s] for removal", $sMyName, $sLockFileWithPath );
                   push ( @lsFiles, $sLockFileWithPath );
               }
           }  
           #----------------------
           # Removing Lock files
           #----------------------
           if ( scalar @lsFiles && ! Sys::UnlinkFilesAsUser ( @lsFiles ) )
           {
               &logError ( 20017, "<%s> Failed to remove stale file(s) [%s]",
                                   $sMyName, join ( ' ', @lsFiles ) );
               $iReturnCode = 0; # Failure..
           }
       }
   }
  my $wlsFlag = 0;
  my $sStartScript     = $sScriptHome . "/" . $hScripts{$sServerType};
  my $sStartOptions = "start";
  my $iCommandExitCode = 1;
  my @lsCommandOut = ();
  my $bTimedOut    = 0; 
  my $sRunFromDir  = 0;
  my $iTimeOut     = ( $iExpirationTime - time - 10 );
  my $bSystemD     = SystemD::DetectSystemd();
  if ( ( "12.2" eq $::gnOracleAppsVersion ) && (( $sServerType eq "nodemanager" ) || ( $sServerType eq "adminserver" ) || ( $sServerType eq "oacore" ) || ( $sServerType eq "oafm" ) || ( $sServerType eq "formsserver"   && $sFormsMode eq "servlet") ) ) {
      $wlsFlag = 1;
      my $sStartOptions = "start " . $sServerName ;
      my $sOutFile = "/tmp/.VRTS$sResourceName"."_$sEntryPointName.out";
      my $sDecryptedPassword = DecryptPassword($sWLSPassword);
      if  ( ! $sDecryptedPassword )
      {
        &logError(20130,"<%s>:Could not decrypt the password",$sMyName);
         return 0;
      }

      unlink ("$sOutFile");
      my @sCommandArgsList = $sDecryptedPassword . "\n" ;
      if ($sServerType eq "adminserver") {
        my $sDecryptedAppsPassword = DecryptPassword($sAppsPassword);
        if  ( ! $sDecryptedAppsPassword ){
          &logError(20175,"<%s>:Could not decrypt the Apps password",$sMyName);
          return 0;
        }
        push @sCommandArgsList, "$sDecryptedAppsPassword\n";
        $sStartOptions = "start";
      }
  if ( $bSystemD ) {
    my $sCommandName = undef;
    my $sSystemdStartScript = "/opt/VRTSagents/ha/bin/OracleApps/StartInstanceWithSystemd.pl";
    my $sCommandArgs = " -s $sStartScript -o \"$sStartOptions\" -O \"$sOutFile\" -a \"$sWLSPassword $sAppsPassword\" -u $sUser -r $sResourceName";
 
    if ( $iReturnCode = SystemD::RunSystemdWithEnv("root", $sSystemdStartScript, $sCommandArgs, $sEnvFile, $iTimeOut, $sCommandName ) ) {
      &logTrace ( 20380,
      "<%s> Start command [%s] with args [%s] executed successfully",
      $sMyName, $sStartScript, $sStartOptions );
    }
    else {
      &logError ( 20379, "<%s>: Start command [%s] with args [%s] failed to execute",
          $sMyName, $sStartScript, $sStartOptions);
    }

  }
else {
      $iReturnCode = RunCmdViaPipe ($sUser,
                                    $sStartScript,
                                    $sStartOptions,
                                    $sOutFile,
                                    \@sCommandArgsList,
                                    $sScriptHome,
                                    \$iCommandExitCode,
                                    $iTimeOut,
                                    \$bTimedOut);
}
    }
    else {
  if ( $bSystemD ) {
    if ( $iReturnCode = SystemD::RunSystemdWithEnv($sUser, $sStartScript, $sStartOptions, $sEnvFile, $iTimeOut, undef)) {
       &logTrace ( 20381,
      "<%s> Start command [%s] with args [%s] executed successfully",
      $sMyName, $sStartScript, $sStartOptions );
    }
    else {
      &logError ( 20382, "<%s>: Start command [%s] with args [%s] failed to execute",
          $sMyName, $sStartScript, $sStartOptions );
    }

  }
else {
        $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut( $sEnvFile,
                                                                0,
                                                                $sStartScript,
                                                                $sStartOptions,
                                                                \$iCommandExitCode,
                                                                $sUser,
                                                                \@lsCommandOut,
                                                                \$bTimedOut,
                                                                $iTimeOut,
                                                                "",
                                                                "");
        }
   }
  if ( $iReturnCode )
  {
    #---------------------------------------------------------------------------
    # Command got executed.
    #---------------------------------------------------------------------------
    if ( !$bSystemD && (($wlsFlag == 0 && $iCommandExitCode ) || ($wlsFlag == 1 && $iCommandExitCode != 0 && $iCommandExitCode != 2 ) ) )
    {
      #-------------------------------------------------------------------------
      # Command failed.
      #-------------------------------------------------------------------------
      &logError (20018,"<%s>:[%s] with args [%s] returned exit code [%s]",
                       $sMyName,$sStartScript,$sStartOptions,$iCommandExitCode);
      return 0;
    }
    else
    {
      #-------------------------------------------------------------------------
      # Command executed successfully.
      #-------------------------------------------------------------------------
      &logInfo (20019, "<%s>:[%s] executed successfully.",
                         $sMyName, $sStartScript);
    }
  }
  else
  {
    if ( $bTimedOut )
    {
      #-------------------------------------------------------------------------
      # Command timed out.
      #-------------------------------------------------------------------------
      &logError (20020,"<%s>:[%s] did not complete within [%s] seconds.",
                 $sMyName,$sStartScript,$iTimeOut);
      return 0;
    }
    else
    {
      #-------------------------------------------------------------------------
      # Command failed.
      #-------------------------------------------------------------------------
      &logError (20021,"<%s>:[%s] failed.", $sMyName,$sStartScript);
      return 0;
    }
  }

  #----------------------------------------------------------------
  # Now delay the entry point if we have time to spare, so we 
  # can guarantee that when first monitor is called, it will pass
  #----------------------------------------------------------------

  my $iSleepTime = 10; # Secs..
  while ( ( $iTimeOut = ( $iExpirationTime - time ) ) > 10  )
  {
      sleep $iSleepTime; # Sec..
      &logInfo (20022,"<%s>:In delay entry point . Remaining time = [%s] .", 
                      $sMyName,$iTimeOut);

      if ( VCSDef_MonitorExit::ONLINE ==
         OracleApps::GetProcState ( $iState,
                                    $iIState,
                                    $sUser,
                                    $sServerType,
                                    $sServerName,
                                    $sWLSPassword,
                                    $iPort,
                                    $sOracleHome,
                                    $sDBConnectString,
                                    $sVirtualHostname ) && (   # Running FFLM..
         $iSecondLevelMonitor  ?                              # If enabled..
         VCSDef_MonitorExit::ONLINE ==                        # Online after..
         OracleApps::SecondLevelMonitor ( $iExpirationTime,   # Running SLM..
                                          $sVirtualHostname,
                                          $sUser,
                                          $iPort,
                                          $sOracleHome,
                                          $sScriptHome,
                                          $sServerType,
                                          $sServerName,
                                          $sWLSPassword,
                                          $sAppsPassword,
                                          $sEnvFile,
                                          $sDBConnectString) : 1 ) ) {
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20023,
      "<%s> Confirmed start-up of this instance of [%s] server",
      $sMyName, $sServerType );

	  return 1;
    }
    else {                          
      &logTrace ( 20024,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec", 
      $sMyName, $iSleepTime );
    } 
  }
  &logWarn ( 20025, "<%s> All processes not up after delay entry point ",
                             $sMyName );
  return 0;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN Integer State,
                        IN Integer IState,
                        IN Integer ExpirationTime,
                        IN String  User,
                        IN String  OracleHome,
                        IN String  EnvFile,
                        IN String  ScriptHome,
                        IN String  ServerType,
                        IN String  ServerName,
                        IN String  WLSPassword,
                        IN String  AppsPassword,
                        IN String  Port,
                        IN String  DBConnectString,
                        IN String  VirtualHostname);

 Description : Stop logic needed for offline entry pt

 Parameters :

 $iState                    VCS State of the resource
 $iIState                   VCS Intermidiate State of the resource
 $iExpirationTime           Amount of time left to execute this procedure
 $sUser                     The User to start/stop the Concurrent Manager
 $sOracleHome               Pathname of the ORACLE_HOME directory
 $sEnvFile                  Full pathname to the env file to setup the env
 $sScriptHome               Pathname of the Scripts directory
 $sServerType               Oracle Application Component
 $sServerName               Name of Oracle Application Component
 $sWLSPassword              Password of Weblogic Application Server
 $sAppsPassword             Password for Apps schema
 $iPort                     Port used by component
 $sDBConnectString          DBConnectString if component is listener
 $sVirtualHostname          Virtual Hostname for this component
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iState,
       $iIState,
       $iExpirationTime,
       $sUser,
       $sOracleHome,
       $sEnvFile,
       $sScriptHome,
       $sServerType,
       $sServerName,
       $sWLSPassword,
       $sAppsPassword,
       $iPort,
       $sDBConnectString,
       $sVirtualHostname) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20026, "Subroutine <%s> called with:
  State                    [%s]
  IState                   [%s]
  User                     [%s]
  OracleHome               [%s]
  ScriptHome               [%s]
  EnvFile                  [%s]
  ServerType               [%s]
  ServerName               [%s]
  WLSPassword              [%s]
  AppsPassword             [%s]",

  $sMyName, $iState, $iIState, $sUser,   $sOracleHome,   
  $sScriptHome, $sEnvFile, $sServerType, $sServerName, $sWLSPassword, $sAppsPassword );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  $sServerType =~ tr/A-Z/a-z/;
  my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  my $wlsFlag = 0;
  my $sStopScript     = $sScriptHome . "/" . $hScripts{$sServerType};
  my $sStopOptions = "stop";
  my $iCommandExitCode = 1;
  my @lsCommandOut = ();
  my $bTimedOut    = 0; 
  my $iTimeOut     = ( $iExpirationTime - time - 10 );
  my $sRunFromDir  = 0;

  if ( ( "12.2" eq $::gnOracleAppsVersion ) && ( ( $sServerType eq "nodemanager" ) || ( $sServerType eq "adminserver")  || ( $sServerType eq "oacore" ) || ( $sServerType eq "oafm" ) || ( $sServerType eq "formsserver"   &&    $sFormsMode eq "servlet") ) ) {
      $wlsFlag=1;
      my $sStopOptions = "stop " . $sServerName ;
      my $sOutFile = "/tmp/.VRTS$sResourceName"."_$sEntryPointName.out";
      my $sDecryptedPassword = DecryptPassword($sWLSPassword);
      if  ( ! $sDecryptedPassword )
      {
        &logError(20131,"<%s>:Could not decrypt the password",$sMyName);
         return 0;
      }

      unlink ("$sOutFile");
      my @sCommandArgsList = ( "$sDecryptedPassword\n" );
      if ($sServerType eq "adminserver") {
        my $sDecryptedAppsPassword = DecryptPassword($sAppsPassword);
        if  ( ! $sDecryptedAppsPassword ) {
          &logError(20176,"<%s>:Could not decrypt the Apps password",$sMyName);
          return 0;
        }

        push @sCommandArgsList, "$sDecryptedAppsPassword\n";
      }
      $iReturnCode = RunCmdViaPipe ($sUser,
                                    $sStopScript,
                                    $sStopOptions,
                                    $sOutFile,
                                    \@sCommandArgsList,
                                    $sScriptHome,
                                    \$iCommandExitCode,
                                    $iTimeOut,
                                    \$bTimedOut);
    }
    else {

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut( $sEnvFile,
                                               0,
                                               $sStopScript,
                                               $sStopOptions,
                                               \$iCommandExitCode,
                                               $sUser,
                                               \@lsCommandOut,
                                               \$bTimedOut,
                                               $iTimeOut,
                                               "",
                                               "") ;

  if ( $iReturnCode )
  {
    #---------------------------------------------------------------------------
    # Command got executed.
    #---------------------------------------------------------------------------
    if ( ( $wlsFlag == 0 && $iCommandExitCode  )|| ($wlsFlag == 1 && $iCommandExitCode != 0 && $iCommandExitCode != 2 ))
    {
      #-------------------------------------------------------------------------
      # Command failed.
      #-------------------------------------------------------------------------
      &logError (20027,"<%s>:[%s] with args [%s] returned exit code [%s]",
                       $sMyName,$sStopScript,$sStopOptions,$iCommandExitCode);
      return 0;
    }
    else
    {
      #-------------------------------------------------------------------------
      # Command executed successfully.
      #-------------------------------------------------------------------------
      &logInfo (20028, "<%s>:[%s] executed successfully.",
                         $sMyName, $sStopScript);
    }
  }
  else
  {
    if ( $bTimedOut )
    {
      #-------------------------------------------------------------------------
      # Command timed out.
      #-------------------------------------------------------------------------
      &logError (20029,"<%s>:[%s] did not complete within [%s] seconds.",
                 $sMyName,$sStopScript,$iTimeOut);
      return 0;
    }
    else
    {
      #-------------------------------------------------------------------------
      # Command failed.
      #-------------------------------------------------------------------------
      &logError (20030,"<%s>:[%s] failed.", $sMyName,$sStopScript);
      return 0;
    }
  }
}
  while ( ( $iTimeOut = ( $iExpirationTime - time ) ) > 20  )
  {
      &logInfo (20031,"<%s>:In delay entry point . Remaining time = [%s] .",
                      $sMyName,$iTimeOut);

       my $iReturnCode = OracleApps::GetProcState ( $iState,
                                                    $iIState,
                                                    $sUser,
                                                    $sServerType,
                                                    $sServerName,
                                                    $sWLSPassword,
                                                    $iPort,
                                                    $sOracleHome,
                                                    $sDBConnectString,
                                                    $sVirtualHostname );

       if( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) {
         &logTrace ( 20032, "<%s> All processes are down ", $sMyName );
         return 1;
       }
       else {
         &logTrace ( 20033, "<%s> All required processes not down yet.
                            Time remaining [%s]", $sMyName, $iTimeOut);
        }
      sleep (5);
  }
  &logWarn ( 20034, "<%s> All processes not down after delay entry point ",
                             $sMyName );

  #---------------------
  # Call kill to cleanup
  #---------------------


  if ( OracleApps::Kill ( VCSDef_CleanReason::OFFLINE_INEFFECTIVE,
                          $iExpirationTime,
                          $sUser,
                          $sOracleHome,
                          $sScriptHome,
                          $sEnvFile,
                          $sServerType,
                          $sServerName,
                          $sWLSPassword ) ) {
      #--------------
      # Kill worked..
      #--------------
      &logInfo ( 20035, "<%s> Cleanup of partial online succeeded for [%s]",
                        $sMyName , $sResourceName);
   }
   else {
      #--------------
      # Kill failed..
      #--------------
      &logError ( 20036, "<%s> Clean failed for [%s] failed.",
                         $sMyName , $sResourceName);
      }

  return 0;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,,
                        IN Integer ExpirationTime,
                        IN String User,
                        IN String OracleHome,
                        IN String ScriptHome,
                        IN String EnvFile,
                        IN String ServerType,
                        IN String ServerName,
                        IN String WLSPassword,
                        IN String AppsPassword );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Reason why clean was called
 $iExpirationTime           Amount of time left to execute this procedure
 $sUser                     The User to start/stop the Concurrent Manager
 $sOracleHome               Pathname of the ORACLE_HOME directory
 $sScriptHome               Pathname of the Scripts directory
 $sEnvFile                  Full pathname to the env file to setup the env
 $sServerType               Oracle Application Component
 $sServerName               Name of Oracle Application Component
 $sWLSPassword              Password of Weblogic Application Server
 $sAppsPassword             Password for Apps schema
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $iExpirationTime,
       $sUser,
       $sOracleHome,
       $sScriptHome,
       $sEnvFile,
       $sServerType,
       $sServerName,
       $sWLSPassword,
       $sAppsPassword ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20037, "Subroutine <%s> called with:
  User                     [%s]
  ServerType               [%s]
  ServerName               [%s]
  WLSPassword              [%s]
  AppsPassword             [%s]",

  $sMyName,  $sUser,  $sServerType, $sServerName, $sWLSPassword, $sAppsPassword );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  $sServerType =~ tr/A-Z/a-z/;
  my @liServerPids = ();
  my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  if ( ! GetServerPids ( $sUser, $sServerType, $sOracleHome,\@liServerPids ) )
  {
    #--------------
    # No pids found
    #--------------
    return 1;
  }

  #------------------------------------------------------------------
  # We found pids, get a list for the server type first and then kill
  #-------------------------------------------------------------------
  if (  VCSDef_CleanReason::OFFLINE_HUNG == $iCleanReason ||
        VCSDef_CleanReason::OFFLINE_INEFFECTIVE == $iCleanReason)
  {
      #---------------------------
      # Kill the server process.
      #---------------------------
      &logInfo(20038,"<%s>:Offline was unsuccessful. Killing process for ServerType [%s].",
                        $sMyName , $sServerType );
  }
  else 
  {
      #---------------------------
      # First attempt to stop
      #---------------------------

      &logInfo(20039,"<%s>:Trying a graceful shutdown for ServerType [%s] before killing",
                        $sMyName , $sServerType );
      my $sStopScript     = $sScriptHome . "/" . $hScripts{$sServerType};
      my $sStopOptions = "stop";
      my $iCommandExitCode = 1;
      my @lsCommandOut = ();
      my $bTimedOut    = 0;
      my $iTimeOut     = 10;
      my $sRunFromDir  = 0;
      if ( "12.2" eq $::gnOracleAppsVersion && ( ( $sServerType eq "nodemanager" ) || ( $sServerType eq "adminserver" ) || $sServerType eq "oacore" || $sServerType eq "oafm" || ( $sServerType eq "formsserver"   &&    $sFormsMode eq "servlet") ) ) {
        my $sStopOptions = "stop " . $sServerName ;
        my $sOutFile = "/tmp/.VRTS$sResourceName"."_$sEntryPointName.out";
        my $sDecryptedPassword = DecryptPassword($sWLSPassword);
        if  ( ! $sDecryptedPassword )
          {
          &logError(20132,"<%s>:Could not decrypt the password",$sMyName);
          return 0;
          }

	my @sCommandArgsList = ( "$sDecryptedPassword\n" );
        if ($sServerType eq "adminserver") {
          my $sDecryptedAppsPassword = DecryptPassword($sAppsPassword);
          if  ( ! $sDecryptedAppsPassword ) {
            &logError(20177,"<%s>:Could not decrypt the Apps password",$sMyName);
            return 0;
          }
          push @sCommandArgsList, "$sDecryptedAppsPassword\n";
        }

       unlink ("$sOutFile");
        $iReturnCode = RunCmdViaPipe ($sUser,
                                    $sStopScript,
                                    $sStopOptions,
                                    $sOutFile,
                                    \@sCommandArgsList,
                                    $sScriptHome,
                                    \$iCommandExitCode,
                                    $iTimeOut,
                                    \$bTimedOut);

      }
      else {

        Sys::RunWithEnvCmdWithOutputWithTimeOut( $sEnvFile,
                                               0,
                                               $sStopScript,
                                               $sStopOptions,
                                               \$iCommandExitCode,
                                               $sUser,
                                               \@lsCommandOut,
                                               \$bTimedOut,
                                               $iTimeOut,
                                               "",
                                               "") ;

      
      }
    }
    #-----------------------------------------------------------------------------
    # We don't care whether stop succeeded or not, we will just kill
    #-----------------------------------------------------------------------------
    my $blongproclist = 1;
    if (Proc::KillPidsFromList(\@liServerPids,$blongproclist))
    {
      &logTrace(20040,"<%s> Killed process for Server [%s] ",
                       $sMyName, $sServerType);
    }
    else
    {
      &logError(20041,"<%s>:Error killing processes for Server [%s]",
                       $sMyName, $sServerType);
      $iReturnCode = 0;

    }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetServerPids ( IN String User,
                                IN String ServerType,
                                IN String OracleHome,
                                OUT ListReference CompPids );

 Description : Get the list of pids for a server type

 Parameters :

 $sUser               User running the component
 $sServerType         Oracle Application Component
 $sOracleHome         Pathname of the ORACLE_HOME directory
 $rliCompPids         Ref. to list for online PIDs


 Returns : 1 found pids
           0 no pids found

           Populate the list with the pids

 -------------------------------------------------------------------------

=cut

sub GetServerPids
{
  my ( $sUser,
       $sServerType,
       $sOracleHome,
       $rliCompPids ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20042, "Subroutine <%s> called with:
  User                     [%s]
  ServerType               [%s]
  OracleHome               [%s]
  rliCompPids              [%s]",

  $sMyName,  $sUser,  $sServerType, $sOracleHome, $rliCompPids );

  my @hProcListHash        = ();
  my $iProcessCount = Proc::GetLongProcessListHash(\@hProcListHash);

  if ( 0 == $iProcessCount )
  {
    #----------------------
    # No processes found.
    #----------------------
    &logError(20043,"<%s> : Error getting all running processes.",$sMyName);
     return 0;
  }

  my @hUidProcListHash = ();
  my $iUid = getpwnam $sUser;
  my $iMatchedPids = Proc::FindUidPids ( $iUid,
                                         \@hProcListHash,
                                         \@hUidProcListHash );

  &logTrace(20044,"<%s> : Number of processes for User %s = %s",
                           $sMyName, $sUser, $iMatchedPids);


  $sServerType =~ tr/A-Z/a-z/;

  SWITCH: {
    if ( $sServerType eq "formsmetricsserver" ) {
      @MONITOR_PROC_LIST = @FORMSMETRICSSERVERPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "formsmetricsclient" ) {
      @MONITOR_PROC_LIST = @FORMSMETRICSCLIENTPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "formsserver" ) {
      @MONITOR_PROC_LIST = @FORMSSERVERPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "listener" ) {
      @MONITOR_PROC_LIST = @LISTENERPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "fulfillmentserver" ) {
      @MONITOR_PROC_LIST = @FULFILLMENTSERVERPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "reportserver" ) {
      @MONITOR_PROC_LIST = @REPORTSERVERPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "webserver" ) {
      @MONITOR_PROC_LIST = @WEBSERVERPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "discoverer" ) {
      @MONITOR_PROC_LIST = @DISCOVERERPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "oacore" ) {
      @MONITOR_PROC_LIST = @OACOREPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "oafm" ) {
      @MONITOR_PROC_LIST = @OAFMPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "opmn" ) {
      @MONITOR_PROC_LIST = @OPMNPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "adminserver" ) {
      @MONITOR_PROC_LIST = @ADMINSERVERPROC_LIST;
      last SWITCH;
    }
    if ( $sServerType eq "nodemanager" ) {
      @MONITOR_PROC_LIST = @NODEMANAGERPROC_LIST;
      last SWITCH;
    }
  }

  my $iFoundProcs = 0 ;
  my $sFilter = "";
  my @hProcListHashMatched = ();

  my $iFound = 0 ;
  my $iNumberOfMatches = 0 ;

  foreach my $sProc (@MONITOR_PROC_LIST) 
  {
    $sFilter = "^.*" . $sProc . ".*" . "\\b" ;

    if ( $sServerType eq "formsserver" && $sProc eq "f60webmx" ) {
      if ( $^O eq "aix" )
      {
        next;
      }
      $sFilter = "^$sProc" . "*" . "\\b" ;
    }

    &logTrace(20045,"<%s> : Filter [%s] for Server [%s] ",
                          $sMyName, $sFilter, $sServerType);

    $iNumberOfMatches = Proc::FilterProcs($sFilter, 
                                        \@hUidProcListHash, 
                                        \@hProcListHashMatched);


    if ( 0 == $iNumberOfMatches )
    {
      &logTrace(20046,"<%s>:No processes found component.",$sMyName);
    }
    else
    {
      $iFound++;
    }

    if ( $iFound ) 
    {
      my @liStartPidList;
      my $iMatchedProcCount = Proc::GetValuesForKeyFromProcDataHash( "PID", 
                                                                 \@hProcListHashMatched, 
                                                                 \@liStartPidList);
       if ( $iMatchedProcCount >= 1)
       {
         &logTrace(20047,"<%s>:Multiple processes [%s] running server [%s] .",
                          $sMyName,$iMatchedProcCount, $sServerType);

         #------------------------------------------------------
         # Take all PIDs of the script and find all the children
         # of the script to kill them.
         #------------------------------------------------------

         foreach my $sEachPid (@liStartPidList)
         {
            #-----------------------------------------------------------------------
            # Push the Pids on the list
            #-----------------------------------------------------------------------
            &logTrace(20048,"<%s> : Pid [%s]",$sMyName, $sEachPid);
            push ( @$rliCompPids,$sEachPid);
         }
       }
    }
  }

  return $iFound;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String State,
                                   IN String IState,
                                   IN String User,
                                   IN String ServerType,
                                   IN String ServerName,
                                   IN String WLSPassword,
                                   IN Integer Port,
                                   IN String OracleHome,
                                   IN String DBConnectString,
                                   IN String VirtualHostname );

 Description : First level monitor to determine state of resource.

 Parameters :

 $iState                    VCS State of the resource
 $iIState                   VCS Intermidiate State of the resource
 $sUser                     Owner of this application component 
 $sServerType               Oracle Application Component
 $sServerName               Name of Oracle Application Component
 $sWLSPassword              Password of Weblogic Application Server
 $iPort                     Port used by the component
 $sOracleHome               OracleHome for this component
 $sDBConnectString          Connect string ( if component is Listener)
 $sVirtualHostname          Virtual hostname associated with this component
 

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iState,
       $iIState,
       $sUser,
       $sServerType,
       $sServerName,
       $sWLSPassword,
       $iPort,
       $sOracleHome,
       $sDBConnectString,
       $sVirtualHostname ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20049, "Subroutine <%s> called:
  User                     [%s]
  ServerType               [%s]
  ServerName               [%s]
  WLSPassword              [%s]
  Port                     [%s]
  DBConnectString          [%s]
  VirtualHostname          [%s]",

  $sMyName, $sUser, $sServerType, $sServerName, $sWLSPassword, $iPort, $sDBConnectString, $sVirtualHostname );

  #------------------------------------------------------
  # Do an IP check. This  is a quick check on the offline
  # node
  #------------------------------------------------------

  if (  Comms::IsIPLocal ($sVirtualHostname) ) {
    &logTrace ( 20050, "<%s> plumbed locally [%s]",
                $sMyName, $sVirtualHostname);
  }
  else {
    &logTrace ( 20051, "<%s> not plumbed locally [%s]",
                $sMyName, $sVirtualHostname);
     return ( VCSDef_MonitorExit::OFFLINE );
  }

  my @liDummyPids = ();
  my $iUid = getpwnam $sUser; # Has to succeed..

  my $iReturnCode = OracleApps::GetProcState ( $iState,
                                               $iIState,
                                               $sUser,
                                               $sServerType,
                                               $sServerName,
                                               $sWLSPassword,
                                               $iPort,
                                               $sOracleHome,
                                               $sDBConnectString,
                                               $sVirtualHostname );

  if( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20052,
    "<%s> found all required processes",
    $sMyName );
  }
  else {
    &logTrace ( 20053,
    "<%s> did not find all required processes, returning [%s]",
    $sMyName, $iReturnCode );
  }

  &logTrace ( 20054, "<%s> returning [%s]", $sMyName, $iReturnCode );

  return ( $iReturnCode );
}



=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( IN Integer State,
                               IN Integer IState,
                               IN String User,
                               IN String ServerType,
                               IN String ServerName,
                               IN String WLSPassword,
                               IN Integer Port,
                               IN String OracleHome,
                               IN String DBConnectString,
                               IN String VirtualHostname );

 Description :  Check the state of the processes for the OracleApps Component

 Parameters :

 $iState              VCS State of the resource
 $iIState             VCS Intermidiate State of the resource
 $sUser               Oracle Application User
 $sServerType         Oracle Application Component type
 $sServerName         Name of Oracle Application Component
 $sWLSPassword        Password of Weblogic Application Server
 $iPort               Port used by the component
 $sOracleHome         OracleHome for this component
 $sDBConnectString    Connect string ( if component is Listener)
 $sVirtualHostname    Virtual hostname associated with this component
 

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {

  #----------------
  # Get Arguments..
  #----------------

  my ( $iState,
       $iIState,
       $sUser,
       $sServerType,
       $sServerName,
       $sWLSPassword,
       $iPort,
       $sOracleHome,
       $sDBConnectString,
       $sVirtualHostname ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20055, "Subroutine <%s> called:
  User       [%s]
  ServerType [%s]
  ServerName [%s]
  WLSPassword[%s]
  Port       [%s]
  OracleHome [%s]
  DBConnectString [%s]
  VirtualHostname [%s]",

  $sMyName, $sUser, $sServerType, $sServerName, $sWLSPassword, 
  $iPort, $sOracleHome, $sDBConnectString, $sVirtualHostname);

  my @hProcListHash = ();
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;
  my $iNumberOfMatches = 0;

  my $iUid = getpwnam $sUser;
  my $iProcessCount = Proc::GetLongProcessListHash(\@hProcListHash);

  if ( 0 == $iProcessCount )
  {
    #--------------------------------------------------------------------------
    # No processes found.
    #--------------------------------------------------------------------------
    &logError(20056,"<%s> : Error getting all running processes.",$sMyName);
    return ( $VCSDef_MonitorExit::OFFLINE);
  }

  my @hUidProcListHash = ();
  my $iMatchedPids = Proc::FindUidPids ( $iUid,
                                         \@hProcListHash,
                                         \@hUidProcListHash );

  &logTrace(20057,"<%s> : Number of processes for User %s = %s",
                           $sMyName, $sUser, $iMatchedPids);

 #--------------------------------
 # Convert Servertype to lowercase 
 #--------------------------------

 my $iExpectedProcs = 0 ;
 $sServerType =~ tr/A-Z/a-z/;

 SWITCH: {
    if ( $sServerType eq "formsmetricsserver" ) {
      @MONITOR_PROC_LIST = @FORMSMETRICSSERVERPROC_LIST;
      $iExpectedProcs = 1 ;
      last SWITCH;
    }
    if ( $sServerType eq "formsmetricsclient" ) {
      @MONITOR_PROC_LIST = @FORMSMETRICSCLIENTPROC_LIST;
      $iExpectedProcs = 1 ;
      last SWITCH;
    }
    if ( $sServerType eq "formsserver" ) {
      @MONITOR_PROC_LIST = @FORMSSERVERPROC_LIST;
      if ($::gnOracleAppsVersion eq "11" ) {
          $iExpectedProcs = 2 ;
          if ( $^O eq "aix" ) {
              $iExpectedProcs = 1 ;
          }
      }
      else {
          $iExpectedProcs = 1 ;
      }
      last SWITCH;
    }
    if ( $sServerType eq "listener" ) {
      @MONITOR_PROC_LIST = @LISTENERPROC_LIST;
      $iExpectedProcs = 1 ;
      last SWITCH;
    }
    if ( $sServerType eq "reportserver" ) {
      @MONITOR_PROC_LIST = @REPORTSERVERPROC_LIST;
      $iExpectedProcs = 1 ;
      last SWITCH;
    }
    if ( $sServerType eq "webserver" ) {
      @MONITOR_PROC_LIST = @WEBSERVERPROC_LIST;
      if ($::gnOracleAppsVersion eq "11" ) {
          $iExpectedProcs = 1 ;
      }
      else {
          $iExpectedProcs = 2 ;
      }
      last SWITCH;
    }
    if ( $sServerType eq "discoverer" ) {
      @MONITOR_PROC_LIST = @DISCOVERERPROC_LIST;
      $iExpectedProcs = 2;
      last SWITCH;
    }
    if ( $sServerType eq "fulfillmentserver" ) {
      @MONITOR_PROC_LIST = @FULFILLMENTSERVERPROC_LIST;
      $iExpectedProcs = 1;
      last SWITCH;
    }
    if ( $sServerType eq "oacore" ) {
      @MONITOR_PROC_LIST = @OACOREPROC_LIST;
      $iExpectedProcs = 1;
      last SWITCH;
    }
    if ( $sServerType eq "oafm" ) {
      @MONITOR_PROC_LIST = @OAFMPROC_LIST;
      $iExpectedProcs = 1;
      last SWITCH;
    }
    if ( $sServerType eq "opmn" ) {
      @MONITOR_PROC_LIST = @OPMNPROC_LIST;
      $iExpectedProcs = 1;
      last SWITCH;
    }
    if ( $sServerType eq "adminserver" ) {
      @MONITOR_PROC_LIST = @ADMINSERVERPROC_LIST;
      $iExpectedProcs = 1;
      last SWITCH;
    }
    if ( $sServerType eq "nodemanager" ) {
      @MONITOR_PROC_LIST = @NODEMANAGERPROC_LIST;
      $iExpectedProcs = 1;
      last SWITCH;
    }
   
  }
 
 my $iFoundProcs = 0 ;
 my $sFilter = "";

  foreach my $sProc (@MONITOR_PROC_LIST) {
    my @hProcListHashMatched = ();
   
    $sFilter = "^.*" . $sProc . ".*" . "\\b" ;

    if ( $sServerType eq "formsserver" && $sProc eq "f60webmx" ) {
      if ( $^O eq "aix" )
      {
        next;
      }
      $sFilter = "^$sProc" . "*" . "\\b" ;
    }


    &logTrace(20058,"<%s> : Filter [%s] for Server [%s] ",
                            $sMyName, $sFilter, $sServerType);

    $iNumberOfMatches = Proc::FilterProcs($sFilter, 
                                          \@hUidProcListHash, 
                                          \@hProcListHashMatched);


    if ( 0 >= $iNumberOfMatches )
    {
       &logTrace(20059,"<%s> : Did not find process [%s]", $sMyName, $sProc);
    }
    else
    {
       &logTrace(20060,"<%s> : Found match for <%s> ",$sMyName, $sFilter);
       $iFoundProcs++;
    }
  }

  if ( 0 == $iFoundProcs  ) {
       &logTrace(20061,"<%s> : Did not find any processes for [%s]", $sMyName, $sServerType);
       return VCSDef_MonitorExit::OFFLINE;
  }
  elsif ( $iFoundProcs < $iExpectedProcs ) {
       &logTrace(20062,"<%s> : Did not find all processes for [%s]", $sMyName, $sServerType);
       return VCSDef_MonitorExit::PARTIAL_ONLINE;
  }
  else {
       &logTrace(20063,"<%s> : Found all processes for [%s]", $sMyName, $sServerType);
       return VCSDef_MonitorExit::ONLINE;
  }
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN State,
                          IN IState,
                          IN ExpirationTime
                          IN User
                          IN ServerType
                          IN ServerName
                          IN WLSPassword
                          IN AppsPassword
                          IN Port
                          IN OracleHome
                          IN ScriptHome
                          IN EnvFile
                          IN DBConnectString
                          IN SecondLevelMonitor,
                          IN MonitorProgramAndArgs,
                          IN VirtualHostname );


 Description: Performs the detailed monitoring for each component

 Parameters :
 
 $iState                    VCS State of the resource
 $iIState                   VCS Intermidiate State of the resource
 $iExpirationTime           Time left to complete this procedure
 $sUser                     The User to start/stop the Concurrent Manager
 $sServerType               Oracle Application Component
 $sServerName               Name of Oracle Application Component
 $sWLSPassword              Password of Weblogic Application Server
 $sAppsPassword             Password for Apps schema
 $iPort                     Port Number the component is listening on
 $sOracleHome               Pathname of the ORACLE_HOME directory
 $sScriptHome               Pathname of the Oracle Start/Stop Scripts Dir
 $sEnvFile                  Full pathname to the env file to setup the env
 $sDBConnectString          The TNS connect string to ping the listener
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 $sMonitorProgramAndArgs    Full path to external monitor program 
 $sVirtualHostname          Virtual hostname associated with this CM instance

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iState,
       $iIState,
       $iExpirationTime,
       $sUser,
       $sServerType,
       $sServerName,
       $sWLSPassword,
       $sAppsPassword,
       $iPort,
       $sOracleHome,
       $sScriptHome,
       $sEnvFile,
       $sDBConnectString,
       $iSecondLevelMonitor,
       $sMonitorProgramAndArgs,
       $sVirtualHostname ) = @_ ;


  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20064, "Subroutine <%s> called with:
  State                 [%s]
  IState                [%s]
  ExpirationTime        [%s]
  User                  [%s]
  ServerType            [%s]
  ServerName            [%s]
  WLSPassword           [%s]
  AppsPassword          [%s]
  Port                  [%s]
  OracleHome            [%s]
  ScriptHome            [%s]
  EnvFile               [%s]
  DBConnectString       [%s]
  SecondLevelMonitor    [%s]
  MonitorProgramAndArgs [%s]
  VirtualHostname       [%s]",

  $sMyName, $iState, $iIState, $iExpirationTime, $sUser, $sServerType, $sServerName, $sWLSPassword,
  $sAppsPassword, $iPort, $sOracleHome, $sScriptHome, $sEnvFile, $sDBConnectString , $iSecondLevelMonitor,
  $sMonitorProgramAndArgs, $sVirtualHostname);

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  my $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };

  if ( OracleApps::RunSecondLevelMonitor ( $iSecondLevelMonitor, 
                                           $sResourceType,
                                           $sResourceName ) ) 
  {
     $iReturnCode = OracleApps::SecondLevelMonitor($iExpirationTime,
                                                   $sVirtualHostname,
                                                   $sUser,
                                                   $iPort,
                                                   $sOracleHome,
                                                   $sScriptHome,
                                                   $sServerType,
                                                   $sServerName,
                                                   $sWLSPassword,
						   $sAppsPassword,
                                                   $sEnvFile,
                                                   $sDBConnectString);

     if ( VCSDef_MonitorExit::ONLINE == $iReturnCode )  {
       &logTrace ( 20065, "<%s> : Resource [%s] online after SLM ",
                           $sMyName, $sResourceName );
     }
     else {
       &logTrace ( 20066, "<%s> : Resource [%s] not online after SLM . Status = [%s]",
                           $sMyName, $sResourceName , $iReturnCode);
       return VCSDef_MonitorExit::OFFLINE;
     }
 
  }

  #--------------------------------------------------------
  # We are done monitoring if MonitorProgram is not defined
  #--------------------------------------------------------
  if ( $sMonitorProgramAndArgs eq "" ) {
    &logTrace ( 20067, "<%s> : MonitorProg is Null ", $sMyName);
    return $iReturnCode;
  }


  #--------------------------------------------------
  # Run external user provided monitor program
  #--------------------------------------------------
  my ($sMonitorProgram,$sMonitorProgramArgs) =
                           split(/\s/,$sMonitorProgramAndArgs,2);

  $sMonitorProgramArgs = "" if ( ! defined $sMonitorProgramArgs);

  if ( ! -e $sMonitorProgram ) 
  {
      &logError(20068, "<%s>: MonitorProgram [%s] not executable. 
                              Check existence and permissions.", $sMyName, $sMonitorProgram);

  }

  $iReturnCode = OracleApps::ExternalMonitor($iExpirationTime,
                                             $sUser, 
                                             $sEnvFile,
                                             $sMonitorProgram,
                                             $sMonitorProgramArgs,
                                             $sOracleHome);

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode )  {
     &logTrace ( 20069, "<%s> : Resource [%s] online after external monitor ",
                         $sMyName, $sResourceName );
  }
  else {
    &logTrace ( 20070, "<%s> : Resource [%s] not online after external monitor. Status = [%s]",
                        $sMyName, $sResourceName , $iReturnCode);
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int RunSecondLevelMonitor ( IN $SecondLevelMonitor,
                                        IN $ResourceType,
                                        IN $ResourceName );

 Description: Checks whether time to run SecondLevelMonitor and initializes
              counter

 Parameters :

 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 $sResourceType             VCS resource type
 $sResourceName             VCS resource name

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub RunSecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iSecondLevelMonitor,
       $sResourceType,
       $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20071, "Subroutine <%s> called:
  SecondLevelMonitor        [%s]
  ResourceType              [%s]
  ResourceName              [%s]",

  $sMyName, $iSecondLevelMonitor, $sResourceType, $sResourceName );

  #---------------------------
  # Set Return Code to false..
  #---------------------------
  my $iReturnCode = 0; # Do not run SLM..

  #------------------------
  # Setup local variables..
  #------------------------
  my $sSLMDir = "$VCSLOG/log/tmp/.VRTS$sResourceType";
  my $sSLMFile = "$sSLMDir/$sResourceName.slm";
  my $iSLMCntr = undef;

  #--------------------------------------------
  # Validate SecondLevelMonitor trivial cases..
  #--------------------------------------------
  if ( 0 == $iSecondLevelMonitor ||
       1 == $iSecondLevelMonitor ) {
    $iReturnCode = $iSecondLevelMonitor; # Assign..

    &logTrace ( 20072,
    "<%s> Got request to remove SLM file [%s]. Returning with [%s]",
    $sMyName, $sSLMFile, $iSecondLevelMonitor );

    unlink $sSLMFile if ( -e $sSLMFile ); # Remove SLM file..

    return $iReturnCode;
  }

 #----------------------------------
  # Create history folder if needed..
  #----------------------------------
  if ( -d $sSLMDir ) {
    &logTrace ( 20073, "<%s> SLM folder [%s] exists",
    $sMyName, $sSLMDir );
  }
  else {
    mkdir ( $sSLMDir, 0700 ) ||
    &logWarn ( 20074, "<%s> Failed to create SLM folder [%s]",
    $sMyName, $sSLMDir );
    return 1; # Run SLM..
  }

  #--------------------------
  # Create SLM history file..
  #--------------------------
  if ( sysopen ( SLMFILE, $sSLMFile, O_RDWR|O_SYNC|O_CREAT, 0600 ) ) {
    &logTrace ( 20075,
    "<%s> Opened SLM file [%s] successfully for read/write",
    $sMyName, $sSLMFile );
  }
  else {
    &logWarn ( 20076,
    "<%s> Failed to open SLM file [%s] for read/write",
    $sMyName, $sSLMFile );
    return 1; # Run SLM..
  }

  #---------------
  # Read in file..
  #---------------
  ( $iSLMCntr ) = grep { s/\n// } <SLMFILE>;

  #-----------------
  # Corrupted file..
  #-----------------
  if ( ! defined $iSLMCntr || $iSLMCntr < 0 ) {

    $iSLMCntr ||= ''; # Avoid undefs..

    &logTrace ( 20077,
    "<%s> SLM file [%s] contains [%s]. Re-setting to [%s]",
    $sMyName, $sSLMFile, $iSLMCntr, $iSecondLevelMonitor );
    $iSLMCntr = $iSecondLevelMonitor; # Set to default..
  }

 #-----------------------------------------
  # Update counter, if less than requested..
  #-----------------------------------------
  if ( $iSecondLevelMonitor > $iSLMCntr ) {
    ++ $iSLMCntr;
  }
  else {
    $iSLMCntr = 1;    # Now's the time..
    $iReturnCode = 1; # Run SLM..
  }

  #------------------
  # Update the file..
  #------------------
  seek ( SLMFILE, 0, 0 ); # Rewind before update..

  &logTrace ( 20078,
  "<%s> Updating SLM file [%s] with [%s]. Return code is [%s]",
  $sMyName, $sSLMFile, $iSLMCntr, $iReturnCode );

  close ( SLMFILE ) ||
  &logWarn ( 20079,
  "<%s> Close failed on SLM file [%s] after update",
  $sMyName, $sSLMFile, $! );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN ExpirationTime,
                                     IN String Host,
                                     IN String User,
                                     IN Integer Port,
                                     IN String OracleHome,
                                     IN String ScriptHome,
                                     IN String ServerType,
                                     IN String ServerName,
                                     IN String WLSPassword,
                                     IN String AppsPassword,
                                     IN String EnvFile,
                                     IN String DBConnectString)

 Description: Runs the SecondLevel Monitor by doing to socket connection
              to the port of the component or checking the lsnr via tnsping

 Parameters :

 $iExpirationTime           Time left to complete this procedure
 $sHost                     Virtual hostname associated with this CM instance
 $sUser                     The User to start/stop the Concurrent Manager
 $iPort                     Port Number the component is listening on
 $sOracleHome               Pathname of the ORACLE_HOME directory
 $sScriptHome               Pathname of the Oracle Start/Stop Script Dir
 $sServerType               Oracle Application Component
 $sServerName               Name of Oracle Application Component
 $sWLSPassword              Password of Weblogic Application Server
 $sAppsPassword             Password for Apps schema
 $sEnvFile                  Full pathname to the env file to setup the env
 $sDBConnectString          The TNS connect string to ping the listener

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iExpirationTime,
       $sHost,
       $sUser,
       $iPort,
       $sOracleHome,
       $sScriptHome,
       $sServerType,
       $sServerName,
       $sWLSPassword,
       $sAppsPassword,
       $sEnvFile,
       $sDBConnectString ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20080, "Subroutine <%s> called:
  ExpirationTime    [%s]
  Host              [%s]
  Port              [%s]
  OracleHome        [%s]
  ScriptHome        [%s]
  ServerType        [%s]
  ServerName        [%s]
  WLSPassword       [%s]
  AppsPassword      [%s]
  EnvFile           [%s]
  DBConnectString   [%s]",

  $sMyName, $iExpirationTime, $sHost, $iPort, $sOracleHome, $sScriptHome, 
  $sServerType, $sServerName, $sWLSPassword, $sAppsPassword, $sEnvFile, $sDBConnectString );

  #---------------------------
  # Set Return Code to false..
  #---------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # set optimistically to ONLINE
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  #------------------------------------------------
  # Testing Host/Port in case of Socket Connection.
  #------------------------------------------------
  if ( ($::gnOracleAppsVersion eq "11" ) || 
       ($::gnOracleAppsVersion eq "12" && $sServerType eq "formsserver" && $iPort ) ||
       ($::gnOracleAppsVersion eq "12.2" && $sServerType eq "formsserver" && $iPort ))
  {
    if (  Comms::CheckPort($sHost,$iPort) )
    {
      &logInfo(20081, "<%s>: Connection to [%s]:[%s] for [%s] successful.", 
                       $sMyName, $sHost, $iPort , $sServerType);
      $iReturnCode = VCSDef_MonitorExit::ONLINE; 
    }
    else
    {
      &logInfo(20082, "<%s>: Connection to [%s]:[%s] for [%s] not responding.",
                       $sMyName,$sHost,$iPort, $sServerType);
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; 
    }
  }

  #---------------------------
  # Test Other ServerTypes 
  #---------------------------
  $sServerType =~ tr/A-Z/a-z/;
  my $sServerStatusCmd  = undef;
  my $sServerStatusArgs = undef;
  my $sInstHome = "$sScriptHome/../..";
  my $bEnvFileNeeded = 0;
  if ($::gnOracleAppsVersion eq "11" )
  {
      if ( $sServerType eq "listener" )
      {
          $sServerStatusCmd   = "$sOracleHome/bin/tnsping";
          $sServerStatusArgs  = $sDBConnectString;
      }
  }
  else  ## Testing e-Business R12 Servers
  {
      $bEnvFileNeeded = 1;
      if ( $sServerType eq "listener" && $::gnOracleAppsVersion eq "12" ) ## Testing RPC Listener
      {
           $sServerStatusCmd  = "$sOracleHome/bin/lsnrctl";
           $sServerStatusArgs = "status $sDBConnectString";
      }
      elsif ( ( $sServerType eq "fulfillmentserver" ) || 
              ( $::gnOracleAppsVersion eq "12.2" &&
                ( $sServerType eq "webserver"  ||
                  $sServerType eq "listener"   ||
                  $sServerType eq "opmn"  ) ) )
      {    
           $sServerStatusCmd  = $sScriptHome . "/" . $hScripts{$sServerType};
           $sServerStatusArgs = "status";
      }
      else
      {    if ($::gnOracleAppsVersion eq "12.2"){
             $sInstHome = "$sOracleHome/../..";
             $sServerStatusCmd = "$sInstHome/EBSapps/10.1.2/opmn/bin/opmnctl";
           }
           else {
             $sServerStatusCmd = "$sInstHome/ora/10.1.3/opmn/bin/opmnctl";
           }
           if ( $sServerType eq "opmn" ) ## Testing OPMN Process itself
           {
               $sServerStatusArgs = "status";
           }
           else  ## Testing OPMN Managed Processes.
           {
               $sServerStatusArgs = "status  -noheaders -fmt prt%sta ";
               if ( $sServerType eq "webserver" )
               {
                   $sServerType = "HTTP_Server";
               }
               if ( $sServerType eq "formsserver" )
               {
                   $sServerType = "forms";
               }
           }
       }
  }


    my $iCommandExitCode = 1;
    my @lsCommandOut = ();
    my $bTimedOut    = 0;
    my $iTimeOut     = ( $iExpirationTime - time - 10 );
    my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
    if ( "12.2" eq $::gnOracleAppsVersion &&
       ( $sServerType eq "nodemanager" ||
         $sServerType eq "adminserver" ||
         $sServerType eq "oacore" ||
         $sServerType eq "oafm" ||
         ( ( $sServerType eq "forms" || $sServerType eq "formsserver" ) && $sFormsMode eq "servlet"))) {

        $sServerStatusCmd = $sScriptHome . "/" . $hScripts{$sServerType};
        $sServerStatusArgs = "status " . $sServerName ;
        my $sOutFile = "/tmp/.VRTS$sResourceName"."_$sEntryPointName.out";
        my $sDecryptedPassword = DecryptPassword($sWLSPassword);
        if  ( ! $sDecryptedPassword )
          {
          &logError(20133,"<%s>:Could not decrypt the password",$sMyName);
          return 0;
          }

        my @sCommandArgsList = ( "$sDecryptedPassword\n" );
        if ($sServerType eq "adminserver") {
          my $sDecryptedAppsPassword = DecryptPassword($sAppsPassword);
          if  ( ! $sDecryptedAppsPassword ) {
            &logError(20178,"<%s>:Could not decrypt the Apps password",$sMyName);
            return 0;
          }
          push @sCommandArgsList, "$sDecryptedAppsPassword\n";
        }

        unlink ("$sOutFile"); 
        $iReturnCode = RunCmdViaPipe ($sUser,
                                    $sServerStatusCmd,
                                    $sServerStatusArgs,
                                    $sOutFile,
                                    \@sCommandArgsList,
                                    $sScriptHome,
                                    \$iCommandExitCode,
                                    $iTimeOut,
                                    \$bTimedOut);

  if ( $iReturnCode )
  {
    #---------------------------------------------------------------------------
    # Command got executed.
    #---------------------------------------------------------------------------
    if ( $iCommandExitCode )
    {
      #-------------------------------------------------------------------------
      # Command failed.
      #-------------------------------------------------------------------------
      &logError (20134,"<%s>:[%s] with args [%s] returned exit code [%s]",
                       $sMyName,$sServerStatusCmd,$sServerStatusArgs,$iCommandExitCode);
      return VCSDef_MonitorExit::OFFLINE;
    }
    else
    {
      #-------------------------------------------------------------------------
      # Command executed successfully.
      #-------------------------------------------------------------------------
      &logTrace (20135, "<%s>:[%s] executed successfully.",
                         $sMyName, $sServerStatusCmd);
      if (open ( FH , $sOutFile ) )
      {
        my $sServerNametmp = $sServerName;
        if ( $sServerType eq "adminserver" ){
          $sServerNametmp = "AdminServer";
        }
        elsif( $sServerType eq "nodemanager" ) {
          $sServerNametmp = "Node Manager";
        }

        while (<FH>)
        {
          if( /$sServerNametmp is running/i )
          {
            &logTrace(20136,"<%s>: script reports server running.",$sMyName);
            close FH;
            return VCSDef_MonitorExit::ONLINE;
          }
        }
        &logTrace(20137,"<%s>: script reports server NOT running.",$sMyName);
      }
      else
      {
        &logError(20138,"<%s>:Failed to open SLM logfile [%s]",
                        $sMyName,$sOutFile);
        return VCSDef_MonitorExit::UNKNOWN;
      }

      &logTrace(20139,"<%s>: script reports server not running.",$sMyName);
      return VCSDef_MonitorExit::OFFLINE;

     } ## End Else
   }
 }
  else {

    $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut( $sEnvFile,
                                               $bEnvFileNeeded,
                                               $sServerStatusCmd,
                                               $sServerStatusArgs,
                                               \$iCommandExitCode,
                                               $sUser,
                                               \@lsCommandOut,
                                               \$bTimedOut,
                                               $iTimeOut,
                                               "",
                                               "");

    
  if ( $iReturnCode )
  {
    #---------------------------------------------------------------------------
    # Command got executed.
    #---------------------------------------------------------------------------
    if ( $iCommandExitCode )
    {
      #-------------------------------------------------------------------------
      # Command failed.
      #-------------------------------------------------------------------------
      &logError (20083,"<%s>:[%s] with args [%s] returned exit code [%s]",
                       $sMyName,$sServerStatusCmd,$sServerStatusArgs,$iCommandExitCode);
      return VCSDef_MonitorExit::OFFLINE;
    }
    else
    {
      #-------------------------------------------------------------------------
      # Command executed successfully.
      #-------------------------------------------------------------------------
      &logTrace (20084, "<%s>:[%s] executed successfully.",
                         $sMyName, $sServerStatusCmd);

      if (( $::gnOracleAppsVersion ne "12.2" ) && 
          ( $sServerType eq "opmn" || $sServerType eq "listener" ) )
      {
          return VCSDef_MonitorExit::ONLINE;
	  }
      elsif (( $sServerType eq "webserver" ) && ( $::gnOracleAppsVersion eq "12.2" )) {
          if ( grep { /Alive/ } @lsCommandOut ) {
             return VCSDef_MonitorExit::ONLINE;
          }
          else {
          return VCSDef_MonitorExit::OFFLINE;
          }
      }
      elsif (( $sServerType eq "opmn" ) && ( $::gnOracleAppsVersion eq "12.2" )) {
          if ( grep { /opmn is not running/ } @lsCommandOut ) {
             return VCSDef_MonitorExit::OFFLINE;
          }
          else {
          return VCSDef_MonitorExit::ONLINE;
          }
      }

      elsif ( ( $sServerType eq "listener" ) && ( $::gnOracleAppsVersion eq "12.2" ) ) 
      {
         if ( grep { /STATUS of the LISTENER/ } @lsCommandOut ) {
             return VCSDef_MonitorExit::ONLINE;
          }
          else {
          return VCSDef_MonitorExit::OFFLINE;
          }
        
      }
	  elsif ( $sServerType eq "fulfillmentserver" )
	  {
		  if ( grep { /CONNECTION TO PORT TIMED OUT/ } @lsCommandOut ){
			  return VCSDef_MonitorExit::OFFLINE;
		  }
		  elsif ( grep { /RESPONSE/ } @lsCommandOut ) {
			  return VCSDef_MonitorExit::ONLINE;
		  }
		  else {
			  return VCSDef_MonitorExit::OFFLINE;
		  }
	  }
         
	  else
	  {
          SLMPARSE : foreach my $sOutputLine ( @lsCommandOut )
          {
              $sOutputLine =~ s/ //g;
              last SLMPARSE if ( $sOutputLine eq "\n" );
              (my $sComp,my $sState) = split ( /\|/, $sOutputLine );

              (my $sCompFirst ,my $sComponentName) = split ( /\:/, $sComp );
              if ( $sCompFirst eq "OC4J" )
              {
                 $sComp = $sComponentName;
              }
      
              if ( $sComp eq $sServerType)
              {
                 chomp ( $sState );
                 &logTrace (20085,"<%s>:The State of [%s] is [%s]",
                                 $sMyName,$sComp,$sState);

                 if ($sState eq "Alive" )
                 {    
                     $iReturnCode = VCSDef_MonitorExit::ONLINE;
                 }
                 else
                 {
                     $iReturnCode = VCSDef_MonitorExit::OFFLINE;
                 }
              } 
           } # End SLMPARSE
       } ## End Else
     } ## End Else
  }
  else
  {
    if ( $bTimedOut )
    {
      #-------------------------------------------------------------------------
      # Command timed out.
      #-------------------------------------------------------------------------
      &logError (20086,"<%s>:[%s] did not complete within [%s] seconds.",
                 $sMyName,$sServerStatusCmd,$iTimeOut);
    }
    else
    {
      #-------------------------------------------------------------------------
      # Command failed.
      #-------------------------------------------------------------------------
      &logError (20087,"<%s>:[%s] failed.", $sMyName,$sServerStatusCmd);
    }
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; 
  }
  }
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN ExpirationTime,
                                  String User,
                                  String EnvFile,
                                  String MonitorProgram,
                                  String MonitorArgs,
                                  String OracleHome)

 Description: Runs the External Monitor program defined by the user

 Parameters :

 $iExpirationTime           Time left to complete this procedure
 $sUser                     The User to start/stop the Concurrent Manager
 $sEnvFile                  Full pathname to the env file to setup the env
 $sMonitorProgram           User defined utilty to execute
 $sMonitorArgs              Command line arguments
 $sOracleHome               Pathname of the ORACLE_HOME directory

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iExpirationTime,
       $sUser,
       $sEnvFile,
       $sMonitorProgram,
       $sMonitorArgs,
       $sOracleHome ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20088, "Subroutine <%s> called:
  ExpirationTime    [%s]
  User              [%s]
  EnvFile           [%s]
  MonitorProgram    [%s]
  MonitorArgs       [%s] ",

  $sMyName, $iExpirationTime, $sUser, $sEnvFile, $sMonitorProgram,
  $sMonitorArgs, $sOracleHome );

  #---------------------------
  # Set Return Code to false..
  #---------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # set optimistically to ONLINE

  my $iMonitorProgramTimeout = $iExpirationTime - time - EPOP - EPOP ;

  if ( $iMonitorProgramTimeout < EPOP )
  {
     &logInfo (20089, "<%s> : Not enough time left to run the defined MonitorProgram [%s]", 
                       $sMyName , $sMonitorProgram);
     return $iReturnCode;
  }

  &logTrace(20090,"<%s>:About to run [%s] with arguments [%s]",
                   $sMyName,$sMonitorProgram,$sMonitorArgs);


  my $iMonitorProgramExitCode = 0;
  my @lsCommandOut            = ();
  my $bTimedOut               = 0; 

  if (Sys::RunWithEnvCmdWithOutputWithTimeOut( $sEnvFile,
                                               0,
                                               $sMonitorProgram,
                                               $sMonitorArgs,
                                               \$iMonitorProgramExitCode,
                                               $sUser,
                                               \@lsCommandOut,
                                               \$bTimedOut,
                                               $iMonitorProgramTimeout,
                                               "",
                                               $sOracleHome))
  {
      #--------------------------------------------------------------------------
      # The MonitorProgram was successfully run; check the exit code it returned.
      #--------------------------------------------------------------------------
      if( ( VCSDef_MonitorExit::ONLINE == $iMonitorProgramExitCode ) ||
          ( 0 == $iMonitorProgramExitCode ) )
      {
        $iReturnCode = VCSDef_MonitorExit::ONLINE;
        &logTrace( 20091,"<%s>:MonitorProgram [%s] with args [%s] and user context [%s] 
                          indicated ONLINE (exit code [%s])",
                            $sMyName,
                            $sMonitorProgram,
                            $sMonitorArgs,
                            $sUser,
                            $iMonitorProgramExitCode );

      }
      elsif( ( $iMonitorProgramExitCode == VCSDef_MonitorExit::OFFLINE ) ||
             ( 1 == $iMonitorProgramExitCode ) )
      {
        #----------------------------------------------------------------------
        # We must return PARTIAL_ONLINE to reflect that the first level process
        # check (and possibly the SecondLevelMonitor check) did succeed.
        #----------------------------------------------------------------------

        $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;
        &logError( 20092,"<%s>:MonitorProgram [%s] with args [%s] and user context [%s] 
                          indicated OFFLINE (exit code [%s])",
                            $sMyName,
                            $sMonitorProgram,
                            $sMonitorArgs,
                            $sUser,
                            $iMonitorProgramExitCode);
        VCSagentFW::logOutputFromCommand( "$sMonitorProgram $sMonitorArgs",
                                                  $iMonitorProgramExitCode,
                                                  \@lsCommandOut ,
                                                  LogInt::LLError);
        $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;
      }
      else
      {
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
        &logWarn( 20093, "<%s>:MonitorProgram [%s] with args [%s] and user context [%s] 
                           did not indicate ONLINE nor OFFLINE (it provided exit code [%s])",
                           $sMyName,
                           $sMonitorProgram,
                           $sMonitorArgs,
                           $sUser,
                           $iMonitorProgramExitCode);
        VCSagentFW::logOutputFromCommand( "$sMonitorProgram $sMonitorArgs",
                                                  $iMonitorProgramExitCode,
                                                  \@lsCommandOut,
                                                  LogInt::LLWarn);
      }
    }
    else
    {
      if( $bTimedOut )
      {
        #----------------------------------------------------------------------
        # Command timed out!  Return UNKNOWN as the first (and possibly second)
        # level monitoring did work -- it is considered safer to return UNKNOWN
        # as the service may be generally OK, and the MonitorProgram didn't
        # determine for sure that the service was not up, as the defined
        # methodology to do so (as implemented in the MonitorProgram) could not
        # be run successfully (it timed out).
        #----------------------------------------------------------------------
        &logError( 20094, "<%s>:MonitorProgram [%s] with args [%s] and user context [%s] 
                            failed to run (timed out)",
                            $sMyName,
                            $sMonitorProgram,
                            $sMonitorArgs,
                            $sUser );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }
      else
      {
        #----------------------------------------------------------------------
        # The MonitorProgram was not successfully run; The call to the
        # MonitorProgram failed (but did NOT timeout).  Return UNKNOWN for the
        # same reasons as explained above.
        #----------------------------------------------------------------------
        &logError( 20095, "<%s>:MonitorProgram [%s] with args [%s] and user context [%s] 
                           failed to run )",
                            $sMyName,
                            $sMonitorProgram,
                            $sMonitorArgs,
                            $sUser );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }
  }
  return $iReturnCode;
}

=pod
 -------------------------------------------------------------------------
 Function : int GetEPTimeout ( IN String EntryPointName
                               IN String ResourceType  );

 Description : Returns the Entry Point Timeout for the EP which we are in

 Parameters :

 $sEntryPointName           The EP which we are currently running
 $sResourceType             VCS resource type

 Returns  : Integer value as defined below:

 -------------------------------------------------------------------------
=cut

sub GetEPTimeout
{
  my ( $sEntryPointName, $sResourceType) = @_;

  #------------------
  # Announce Entry..
  #------------------

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20122, "Subroutine <%s> called with:
  EntryPointName [%s]
  ResourceType   [%s]", 
  $sMyName , $sEntryPointName , $sResourceType );

  my $sObjectName          = "hatype";
  my %hTypeValues;

  my $iEngineTimeout;
  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20123,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
    return ( $iEngineTimeout ); 
  }

  if ( ! VCSagentFW::GetVCSObjectHash($sObjectName,$sResourceType,\%hTypeValues) )
  {
    &logError(20124,"<%s>:GetVCSObjectHash call failed.",$sMyName);
    return 0;
  }

  #-------------------------------
  # Make "monitor" to "Monitor"
  # offline to Offline and so on
  #-------------------------------

  my $sTempEntry =  $sEntryPointName ;
  $sTempEntry    =~ s/(\w)(.*)$/\U$1\L$2/; 

  &logTrace(20125,"<%s>: Returning [%s] value [%s]",$sMyName, $sTempEntry."Timeout",$hTypeValues{$sTempEntry."Timeout"});
  return ( $hTypeValues{$sTempEntry."Timeout"});
}

=pod

 ----------------------------------------------------------------------------
 Function: Boolean RunCmdViaPipe( IN  String    User,
                                 IN  String    Command,
                                 IN  String    OutFile,
                                 IN  String    CommandArgs,
                                 IN  String    FromDir,
                                 OUT Integer   CommandReturnCode ,
                                 IN  Integer   TimeoutSeconds,
                                 OUT Boolean   TimedOut);

 Description:
  This function executes the specified command in an interactive program
  via pipe mechanism as  specified  user. It executes   the command with
  timeout value. The output is dup'd in the OutFile specified.
 Parameters:
  User              : The user name with which the command needs to be executed.
  Command           : Command the needs to be executed.
  OutFileName       : Output of the command will be redirected to this file.
  CommandArgs       : The arguments that needs to be passed to the command.
                       Reference to list of commands. Each member of the list
                       is an eol'd string.
  FromDir           : Command will be executed after doing a chdir to this dir.
  CommandReturnCode : The return code of the command executed.
  TimeoutSeconds    : Maximum time the command can execute before getting
                      signaled out due to timeout.
  TimeOut           : Flag indicating whether the command failed to execute
                      due to timeout.

 Return Value:
  Boolean value from the set of {0,1} where
         0 - command timedout
         1 - command not timedout.
 ----------------------------------------------------------------------------

=cut


sub RunCmdViaPipe
{
  my ( $sUser,
       $sCommand ,
       $sCommandOpts,
       $sOutFile,
       $rsCommandArgs ,
       $sBEA_HOME,
       $riCommandReturnCode ,
       $iTimeoutSeconds,
       $rbTimedOut ) = @_;

  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20140, "Subroutine <%s> called with:
  User                     [%s]
  Command                  [%s]
  CommandOpts              [%s]
  OutFileString            [%s]
  BEA_HOME                 [%s]
  CommandReturnCode        [%s]
  TimeoutSeconds           [%s]",
  $sMyName, $sUser, $sCommand, $sCommandOpts, $sOutFile, $sBEA_HOME, $$riCommandReturnCode, $iTimeoutSeconds);

  if ( ! -x $sCommand )
  {
    &logError(20141,"<%s>:[$sCommand] is not executable. Check file existence and permissions.",$sMyName, $sCommand);
    return 0;
  }
  my $sScript = $sCommand;
  $sCommand = $sCommand . " " . $sCommandOpts ;
  $$rbTimedOut = 1;
  my $iChildPid = open( CMD, "|-");

  $SIG{PIPE} = sub {
                     &logError(20142 ,
                               "<%s>:Broken pipe for [%s]" ,
                               $sMyName, $sCommand);
                     return 0;
                   };

  if ( $iChildPid )
  {
    #------------------
    # Parent process.
    #------------------
    $$rbTimedOut = 0;
    eval
    {
      local $SIG{ALRM} = sub
      {
        #---------------------------
        # Alarm handled for timeout
        #---------------------------
        &logInfo(20143,"<%s>:Command [%s] timed out, going to attempt to kill the blocking child processes",
                 $sMyName, $sCommand);
        $$rbTimedOut = 1;
        if( defined $iChildPid && $iChildPid > 1 )
        {
          Proc::KillPIDandAllChildren( $iChildPid );
        }
        else
        {
          &logInfo(20144,"<%s>:The Child PID was either undefined or less than 1. Therefore we cannot kill the blocking child process",$sMyName);
        }
        return 0;
      };

      #------------------
      # Set the timeout
      #------------------
      alarm $iTimeoutSeconds;

      for (@$rsCommandArgs)
      {
        #--------------------------------
        # Write the commands to the pipe
        #--------------------------------
        print CMD ;
      }
      close ( CMD );
      $$riCommandReturnCode = $? >> 8;
      if ( $$riCommandReturnCode == 0 || $$riCommandReturnCode == 2 )
      {
        #--------------------------------
        # Command execution was good.
        #--------------------------------
        &logTrace(20145,"<%s>:[%s] Command executed successfully. Return code [%s]", $sMyName, $sCommand, $$riCommandReturnCode );
      }
      else
      {
        #------------------------------------------
        # Command execution returned some code.
        #------------------------------------------
        &logError(20146,"<%s>:Command [%s] execution was not successful.[%s]", $sMyName , $sCommand , $$riCommandReturnCode);
      }
      #---------------------
      # cancel the alarm
      #---------------------
      alarm 0;

      #---------------------------
      # command didn't timeout.
      #---------------------------

    }; #end of eval
  }
  elsif ( 0 == $iChildPid )
  {
    #------------------
    # Child process.
    #------------------
    my ( $sName      ,
         $sPassword  ,
         $iUid       ,
         $iGid       ,
         $iQuota     ,
         $sComment   ,
         $sGcos      ,
         $sHome      ,
         $sUserShell ,
         $sExpire    ) = getpwnam($sUser);

    chdir ( $sBEA_HOME );
    #------------------------------------------
    # SetGid and Setuid of the User attribute
    #------------------------------------------
    POSIX::setgid($iGid);

    #--------------
    # Get File Gid
    #--------------

    my $iFileGid = 0;

    $iFileGid = ( stat $sScript ) [5];

    &logTrace(20147,"<%s>: Start script Group owner is [%s].", $sMyName, $iFileGid);

    #----------------------
    # Get members of $fGid
    #----------------------

    my ( $sFileGrpName, $lnGrpMembers,
         @lsGrpMembers, $iMemberUid, $sFileUser ) = ();

    $sFileGrpName = getgrgid ( $iFileGid );
    $lnGrpMembers = ( getgrnam $sFileGrpName ) [3];
    @lsGrpMembers = split (' ', $lnGrpMembers );

    &logTrace(20148,"<%s>:  Start script File's Group members are [%s].", $sMyName, $lnGrpMembers);

    #------------------------------
    # Set $iUid is member of $fGid
    #------------------------------
    foreach $sFileUser ( @lsGrpMembers ){
        $iMemberUid = getpwnam ( $sFileUser );
        if ( $iMemberUid && $iMemberUid == $iUid ){
           $) = "$iGid $iFileGid $)";
           &logTrace(20149,"<%s>:Set Gid to [%s].", $sMyName, $iGid);
        }
    }
    &logTrace(20150,"<%s>:Set Uid to [%s].", $sMyName, $iMemberUid);
    POSIX::setuid($iUid);

    #----------------------------------------------
    # Dup the stdout and stderr of this child to
    # be exec'd to the OutFile
    #----------------------------------------------
    open STDOUT, '>', "$sOutFile" or
    die "<$sMyName>:Can't redirect STDOUT to $sOutFile: $! ";
    open STDERR, ">&STDOUT" or
    die "<$sMyName>:Can't redirect STDERR to $sOutFile: $!";

    #--------------
    # Exec Command
    #--------------
    exec("$sCommand ") ||
    &logError(20151,"<%s>:Cannot execute [%s]",$sMyName, $sCommand);
    return 0;
  }
  else
  {
    #------------------------
    # Cannot fork process !
    #------------------------
    &logError(20152,"<%s>:Cannot fork [%s]",$sMyName, $sCommand);
    return 0;
  }

  if ( $$rbTimedOut )
  {
    &logError(20153,"<%s>:The command did not complete within [%s] seconds",
                    $sMyName, $iTimeoutSeconds );
    return 0; # return invalid as the script timed out.
  }
  else
  {
    return 1;
  }
}

=pod
----------------------------------------------------------------------------
 Function:  String DecryptPassword ( IN String CryptedPassword );

 Description : Decrypts the password using VCS decrypt subroutine.

 Parameters :
  $sCryptedPassword       Decrypts the password

 Returns : None
----------------------------------------------------------------------------

=cut

sub DecryptPassword
{
  my ($sCryptedPassword ) = @_;

  my $sMyName = ( caller 0 )[3];
  &logTrace(20154,"Subroutine <%s> called.", $sMyName );

  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  $sResourceType = 'OracleApps' if $sResourceType !~ /\w/;

  my $iRelevantTimeout = OracleApps::GetEPTimeout($sEntryPointName, $sResourceType);
  $iRelevantTimeout ||= 30; # Avoid undefs.. Default MonitorTimout/2

  my $sDecryptCmd = $HADECRYPT ||= "/opt/VRTSvcs/sbin/vcsdecrypt";

  #-----------------------------------------------------------------
  # Check decrypt command and return immediately if not executable.
  #-----------------------------------------------------------------
  if ( ! -x $sDecryptCmd )
  {
    &logError( 20155, "<%s>:Decrypt command [%s] is not an executable.",
                       $sMyName, $sDecryptCmd);
    return 0;
  }
  #----------------------------------------
  # Execute the actual vcsdecrypt command.
  #----------------------------------------
  my $iExitCode;
  my @lsCommandOut;
  my $bTimedOut,
  my $iTimeOut = ${iRelevantTimeout}/2;
  my $iReturnCode;
  my $bDisplayOutput = 0;  #Dont display the password

  $iReturnCode = OracleApps::__RunWithEnvCmdWithOutputWithTimeOut(
                                                  $sDecryptCmd,
                                                  $sCryptedPassword,
                                                  \$iExitCode,
                                                  "root",
                                                  \@lsCommandOut,
                                                  \$bTimedOut,
                                                  $iTimeOut,
                                                  "",
                                                  "",
                                                  $bDisplayOutput);
  if ( $iReturnCode )
  {
    #----------------------------------
    # Command executed.
    #----------------------------------
    if ( $iExitCode )
    {
      #-----------------
      # Command failed.
      #-----------------
      &logError (20156,"<%s>:[%s] failed with exit code [%s]",
                     $sMyName,$sDecryptCmd,$iExitCode);
      return ;
    }
    else
    {
      #---------------------------------
      # Command executed successfully.
      #---------------------------------
      &logTrace (20157, "<%s>:Command [%s] executed successfully.",
                 $sMyName, $sDecryptCmd);
      return $lsCommandOut[0] ;
    }
  }
  else
  {
    #-----------------------------------
    # Problem during command execution.
    #-----------------------------------
    if ( $bTimedOut )
    {
      #--------------------
      # Command timed out.
      #--------------------
      &logError (20158,"<%s>:[%s] did not complete within [%s] seconds.",
                 $sMyName,$sDecryptCmd,$iTimeOut);
      return;
    }
    else
    {
      #--------------------
      # Command failed.
      #--------------------
      &logError (20159,"<%s>:[%s] failed.", $sMyName,$sDecryptCmd);
    }
    return;
  }
}

=pod
 ----------------------------------------------------------------------------
 Function: Boolean __RunWithEnvCmdWithOutputWithTimeOut(
                                     IN  String    Command,
                                     IN  String    CommandArgs,
                                     OUT Integer   CommandReturnCode ,
                                     IN  String    User,
                                     OUT Array     CommandOut ,
                                     OUT Boolean   TimedOut,
                                     IN  Integer   TimeoutSeconds ,
                                     IN  String    OutFileName ,
                                     IN  String    FromDir );

 Description:
  This function executes the specified command by sourcing environment file
  if needed and as specified user. It executes the command with timeout value.

 Parameters:
  Command           : Command the needs to be executed.
  CommandArgs       : The arguments that needs to be passed to the command.
  CommandReturnCode : The return code of the command executed.
  User              : The user name with which the command needs to be executed.
  CommandOut        : The output of the command executed.
  TimeOut           : Flag indicating whether the command failed to execute
                      due to timeout.
  TimeoutSeconds    : Maximum time the command can execute before getting
                      signaled out due to timeout.
  OutFileName       : Output of the command will be redirected to this file,
                      only if specified. NULL value is permitted.
  FromDir           : Command will be executed after doing a chdir to this dir.
  DisplayOutput     : Boolean Command Output will be displayed if set to 1.

 Return Value:
  Boolean value from the set of {0,1} where
         0 - command timedout
         1 - command not timedout.
----------------------------------------------------------------------------

=cut

sub __RunWithEnvCmdWithOutputWithTimeOut
{
  my ( $sCommand,
       $sCommandArgs,
       $riCommandReturnCode,
       $sUser,
       $rlsCommandOut,
       $rbTimedOut,
       $iTimeoutSeconds,
       $sOutFileName,
       $sFromDir,
       $bDisplayOutput ) = @_ ;

  my $sMyName = (caller 0)[3];

  &logTrace(20160,"<%s>:Command [%s], User [%s], Timeout [%s], OutFile [%s], FromDir [%s], DisplayOutput [%s]",
                   $sMyName, $sCommand, $sUser, $iTimeoutSeconds, $sOutFileName, $sFromDir , $bDisplayOutput );

  #-----------------------------
  # Establish local variables.
  #-----------------------------
  my $sCmdToRun     = "";
  my $sCommandLine  = "";

  if( ! defined $sUser || $sUser eq "" ) {
    &logTrace(20161,"<%s>:No user is set. Setting User to root.", $sMyName );
    $sUser = "root";
  }

  my ( $sName      ,
       $sPassword  ,
       $iUid       ,
       $iGid       ,
       $iQuota     ,
       $sComment   ,
       $sGcos      ,
       $sHome      ,
       $sUserShell ,
       $sExpire    ) = getpwnam($sUser);

  my $bCshSet       = 0;
  if ( $sUserShell =~ /(.*)\/bin\/.*csh$/ )
  {
    $bCshSet = 1;
  }

  #----------------------------------------------------------------
  # Run command from the directory specified.
  #----------------------------------------------------------------

  if( $sFromDir && $sFromDir ne "" ) {
     if ( -d $sFromDir ) {
        &logTrace(20162,"<%s:>Running command from directory [%s]",$sMyName, $sFromDir );
        $sCmdToRun .= "cd $sFromDir ;";
     }
     else
     {
        &logTrace(20163,"<%s:>Directory [%s] does not exist",$sMyName, $sFromDir );
     }
  }

  #------------------------------------
  # Confirm existance of Exectuable.
  #------------------------------------
  if ( ! -e $sCommand ) {
    &logError(20164,"<%s>:File [%s] does not exist",$sMyName, $sCommand ) ;
    return 0 ;
  }
  elsif ( ! -x $sCommand ) {
    &logError(20165,"<%s>:File [%s] is not executable",$sMyName, $sCommand ) ;
    return 0 ;
  }
  else {
    if( defined $sCommandArgs || $sCommandArgs ne "" ) {
      $sCmdToRun .= "$sCommand $sCommandArgs" ;
    }
    else {
      $sCmdToRun .= $sCommand ;
    }
  }

  if ( $sOutFileName )
  {
    if ( $bCshSet )
    {
      $sCmdToRun .= ">& $sOutFileName ";
    }
    else
    {
      $sCmdToRun .= "> $sOutFileName 2>&1";
    }
  }

  &logTrace(20166,"<%s:>Going to run command line [%s] as User [%s]",
                        $sMyName , $sCmdToRun , $sUser );

  #---------------------------------------------------------------------
  # If the request is to run the command within the same user context as
  # this process is already running within, then we do not have to
  # do the command from under "su":
  #---------------------------------------------------------------------
  $sCommandLine = "$::gsCommandSu $sUser $::gsCommandArgsSu '$sCmdToRun'";

  #---------------------------------------
  # Set timeout flag to true for now.
  #---------------------------------------
  $$rbTimedOut = 1;

  if ($iTimeoutSeconds < 1 ) {
    &logTrace(20167,"<%s>:Passed in timeout is less than 1 second, defaulting it to 15 seconds.",$sMyName) ;
    $iTimeoutSeconds = 15;
  }
  my $iChildPid = '';
  my @sCmdOut   = ();

  eval {
    local $SIG{ALRM} = sub {
      &logTrace(20168,"<%s>:Command timed out, going to attempt to kill the blocking child processes",$sMyName);
      if( defined $iChildPid && $iChildPid > 1 ) {
        Proc::KillPIDandAllChildren( $iChildPid );
      }
      else {
        &logTrace(20169,"<%s>:The Child PID was either undefined or less than 1.  Therefore we cannot kill the blocking child process",$sMyName );
      }
      die "SIGALARM JUMP EVAL";
    };
    alarm $iTimeoutSeconds;

    my $sFullCmd = "$sCommandLine 2>&1";
    $iChildPid = open(CMD, "$sFullCmd |")
         or &logError(20170, "<%s>:Can't fork a process: [%s]", $sMyName, $! ) ;
    if( defined $iChildPid && $iChildPid > 1 ) {
      @sCmdOut = <CMD>;
    }
    else {
      &logError(20171,"<%s>:The attempt to open a pipe to command resulted in an invalid child PID of [%s]",
                       $sMyName , $iChildPid );
    }
    my $sLastOS_Error = $!;
    if( close(CMD) ) {
      $$riCommandReturnCode = $? >> 8;
    }
    else {
      $$riCommandReturnCode = $? >> 8;
      #-----------------------------------------------------------------------
      # Commands that exit non-zero, which sometimes does not indicate that
      # it failed, causes the close to return zero instead of >= 1.
      #-----------------------------------------------------------------------
      &logTrace(20172,"<%s>:Command line provided a non-zero exit code -- this does not necessarily indicate a problem... (Perl's OS   error variable prior to the command-pipe close was [%s], and after the close was [%s] )",
                       $sMyName , $sLastOS_Error, $! );
    }

    @$rlsCommandOut = @sCmdOut;

    #---------------------
    # cancel the alarm
    #---------------------
    alarm 0;

    #---------------------------
    # command didn't timeout.
    #---------------------------
    $$rbTimedOut = 0;
  };

  if ( $$rbTimedOut ) {
    &logError(20173,"<%s>:The command did not complete within [%s] seconds",
                     $sMyName , $iTimeoutSeconds );

    return 0; # return invalid as the script timed out.
  }
  else
  {
    return 1;
  }
}


1; # So PERL knows its included..
